# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import allensdk.ephys.ephys_features as ft
import numpy as np
import os
path = os.path.dirname(__file__)


def test_v_and_t_are_arrays():
    v = [0, 1, 2]
    t = [0, 1, 2]
    with pytest.raises(TypeError):
        ft.detect_putative_spikes(v, t)

    with pytest.raises(TypeError):
        ft.detect_putative_spikes(np.array(v), t)


def test_size_mismatch():
    v = np.array([0, 1, 2])
    t = np.array([0, 1])
    with pytest.raises(ft.FeatureError):
        ft.detect_putative_spikes(v, t)


def test_find_time_out_of_bounds():
    t = np.array([0, 1, 2])
    t_0 = 4

    with pytest.raises(ft.FeatureError):
        ft.find_time_index(t, t_0)


def test_dvdt_no_filter():
    t = np.array([0, 1, 2, 3])
    v = np.array([1, 1, 1, 1])

    assert np.allclose(ft.calculate_dvdt(v, t), np.diff(v) / np.diff(t))


def test_fixed_dt():
    t = [0, 1, 2, 3]
    assert ft.has_fixed_dt(t) == True

    # Change the first time point to make time steps inconsistent
    t[0] -= 3.
    assert ft.has_fixed_dt(t) == False


def test_detect_one_spike():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    expected_spikes = np.array([728])

    assert np.allclose(ft.detect_putative_spikes(v[:3000], t[:3000]), expected_spikes)


def test_detect_two_spikes():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    expected_spikes = np.array([728, 3386])

    assert np.allclose(ft.detect_putative_spikes(v, t), expected_spikes)


def test_detect_no_spikes():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = np.zeros_like(t)

    assert len(ft.detect_putative_spikes(v, t)) == 0


def test_detect_no_spike_peaks():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = np.zeros_like(t)
    spikes = np.array([])

    assert len(ft.find_peak_indexes(v, t, spikes)) == 0


def test_detect_two_spike_peaks():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    spikes = np.array([728, 3386])
    expected_peaks = np.array([812, 3478])

    assert np.allclose(ft.find_peak_indexes(v, t, spikes), expected_peaks)


def test_filter_problem_spikes():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    spikes = np.array([728, 3386])
    peaks = np.array([812, 3478])

    new_spikes, new_peaks = ft.filter_putative_spikes(v, t, spikes, peaks)
    assert np.allclose(spikes, new_spikes)
    assert np.allclose(peaks, new_peaks)


def test_filter_no_spikes():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = np.zeros_like(t)
    spikes = np.array([])
    peaks = np.array([])

    new_spikes, new_peaks = ft.filter_putative_spikes(v, t, spikes, peaks)
    assert len(new_spikes) == len(new_peaks) == 0


def test_upstrokes():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    spikes = np.array([728, 3386])
    peaks = np.array([812, 3478])

    expected_upstrokes = np.array([778, 3440])
    assert np.allclose(ft.find_upstroke_indexes(v, t, spikes, peaks), expected_upstrokes)


def test_thresholds():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    upstrokes = np.array([778, 3440])

    expected_thresholds = np.array([725, 3382])
    assert np.allclose(ft.refine_threshold_indexes(v, t, upstrokes), expected_thresholds)


def test_thresholds_cannot_find_target():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    upstrokes = np.array([778, 3440])

    expected = np.array([0, 778])
    assert np.allclose(ft.refine_threshold_indexes(v, t, upstrokes, thresh_frac=-5.0), expected)


def test_check_spikes_and_peaks():
    t = np.arange(0, 30) * 5e-6
    v = np.zeros_like(t)
    spikes = np.array([0, 5])
    peaks = np.array([10, 15])
    upstrokes = np.array([3, 13])

    new_spikes, new_peaks, new_upstrokes, clipped = ft.check_thresholds_and_peaks(v, t, spikes, peaks, upstrokes)
    assert np.allclose(new_spikes, spikes[:-1])
    assert np.allclose(new_peaks, peaks[1:])


def test_troughs():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    spikes = np.array([725, 3382])
    peaks = np.array([812, 3478])

    expected_troughs = np.array([1089, 3741])
    assert np.allclose(ft.find_trough_indexes(v, t, spikes, peaks), expected_troughs)


def test_troughs_with_peak_at_end():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    spikes = np.array([725, 3382])
    peaks = np.array([812, 3478])
    clipped = np.array([False, True])

    troughs = ft.find_trough_indexes(v[:peaks[-1]], t[:peaks[-1]],
                                     spikes, peaks, clipped=clipped)
    assert np.isnan(troughs[-1])


def test_downstrokes():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    peaks = np.array([812, 3478])
    troughs = np.array([1089, 3741])

    expected_downstrokes = np.array([862, 3532])
    assert np.allclose(ft.find_downstroke_indexes(v, t, peaks, troughs), expected_downstrokes)


def test_downstrokes_too_many_troughs():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    peaks = np.array([812, 3478])
    troughs = np.array([1089, 3741, 3999])

    with pytest.raises(ft.FeatureError):
        ft.find_downstroke_indexes(v, t, peaks, troughs)


def test_width_calculation():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    spikes = np.array([725, 3382])
    peaks = np.array([812, 3478])
    troughs = np.array([1089, 3741])

    expected_widths = np.array([0.000545, 0.000585])
    assert np.allclose(ft.find_widths(v, t, spikes, peaks, troughs), expected_widths)


def test_width_calculation_too_many_troughs():
    data = np.loadtxt(os.path.join(path, "data/spike_test_pair.txt"))
    t = data[:, 0]
    v = data[:, 1]
    spikes = np.array([725, 3382])
    peaks = np.array([812, 3478])
    troughs = np.array([1089, 3741, 3999])

    with pytest.raises(ft.FeatureError):
        ft.find_widths(v, t, spikes, peaks, troughs)


@pytest.mark.skipif(True, reason="not implemented")
def test_width_calculation_with_burst():
    # example sp 487663469, sweep 43
    pass



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2014-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
import subprocess as sp
import logging
from .config import Config


def choose_bps_command(command='bps_simple', conf_file=None):
    log = logging.getLogger('allensdk.model.biophys_sim.bps_command')

    log.info("bps command: %s" % (command))

    if conf_file:
        conf_file = os.path.abspath(conf_file)

    if command == 'help':
        print(Config().argparser.parse_args(['--help']))
    elif command == 'nrnivmodl':
        sp.call(['nrnivmodl', 'modfiles'])
    elif command == 'run_simple':
        app_config = Config()
        description = app_config.load(conf_file)
        sys.path.insert(1, description.manifest.get_path('CODE_DIR'))
        (module_name, function_name) = description.data[
            'runs'][0]['main'].split('#')
        run_module(description, module_name, function_name)
    else:
        raise Exception("unknown command %s" % (command))


def run_module(description, module_name, function_name):
    m = __import__(module_name, fromlist=[function_name])
    func = getattr(m, function_name)

    func(description)


# this module is designed to be called from the bps script,
# which may use nrniv which does not pass in command line arguments
# So the configuration file path must be set in an environment variable.
if __name__ == '__main__':
    import sys
    conf_file = None
    argv = sys.argv

    if len(argv) > 1:
        if argv[0] == 'nrniv':
            command = 'run_simple'
        else:
            command = argv[1]
    else:
        command = 'run_simple'

    if len(argv) > 2 and (argv[-1].endswith('.conf') or
                          argv[-1].endswith('.json')):
        conf_file = argv[-1]
    else:
        try:
            conf_file = os.environ['CONF_FILE']
        except:
            pass

    choose_bps_command(command, conf_file)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
from . import json_utilities as ju
from allensdk.api.cache import Cache
from allensdk.api.queries.brain_observatory_api import BrainObservatoryApi
from allensdk.config.manifest_builder import ManifestBuilder
from .brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet
import allensdk.brain_observatory.stimulus_info as stim_info
import six
from dateutil.parser import parse as parse_date



class BrainObservatoryCache(Cache):
    """
    Cache class for storing and accessing data from the Brain Observatory.
    By default, this class will cache any downloaded metadata or files in
    well known locations defined in a manifest file.  This behavior can be
    disabled.

    Attributes
    ----------

    api: BrainObservatoryApi instance
        The object used for making API queries related to the Brain
        Observatory.

    Parameters
    ----------

    cache: boolean
        Whether the class should save results of API queries to locations specified
        in the manifest file.  Queries for files (as opposed to metadata) must have a
        file location.  If caching is disabled, those locations must be specified
        in the function call (e.g. get_ophys_experiment_data(file_name='file.nwb')).

    manifest_file: string
       File name of the manifest to be read.  Default is "brain_observatory_manifest.json".
    """

    EXPERIMENT_CONTAINERS_KEY = 'EXPERIMENT_CONTAINERS'
    EXPERIMENTS_KEY = 'EXPERIMENTS'
    CELL_SPECIMENS_KEY = 'CELL_SPECIMENS'
    EXPERIMENT_DATA_KEY = 'EXPERIMENT_DATA'
    STIMULUS_MAPPINGS_KEY = 'STIMULUS_MAPPINGS'
    MANIFEST_VERSION=None

    def __init__(self, cache=True, manifest_file='brain_observatory_manifest.json', base_uri=None):
        super(BrainObservatoryCache, self).__init__(
            manifest=manifest_file, cache=cache, version=self.MANIFEST_VERSION)
        self.api = BrainObservatoryApi(base_uri=base_uri)

    def get_all_targeted_structures(self):
        """ Return a list of all targeted structures in the data set. """
        containers = self.get_experiment_containers(simple=False)
        targeted_structures = set(
            [c['targeted_structure']['acronym'] for c in containers])
        return sorted(list(targeted_structures))

    def get_all_cre_lines(self):
        """ Return a list of all cre driver lines in the data set. """
        containers = self.get_experiment_containers(simple=False)
        cre_lines = set([_find_specimen_cre_line(c['specimen'])
                         for c in containers])
        return sorted(list(cre_lines))

    def get_all_reporter_lines(self):
        """ Return a list of all reporter lines in the data set. """
        containers = self.get_experiment_containers(simple=False)
        reporter_lines = set([_find_specimen_reporter_line(c['specimen'])
                              for c in containers])
        return sorted(list(reporter_lines))

    def get_all_imaging_depths(self):
        """ Return a list of all imaging depths in the data set. """
        containers = self.get_experiment_containers(simple=False)
        imaging_depths = set([c['imaging_depth'] for c in containers])
        return sorted(list(imaging_depths))

    def get_all_session_types(self):
        """ Return a list of all stimulus sessions in the data set. """
        exps = self.get_ophys_experiments(simple=False)
        names = set([exp['stimulus_name'] for exp in exps])
        return sorted(list(names))

    def get_all_stimuli(self):
        """ Return a list of all stimuli in the data set. """
        return sorted(list(stim_info.all_stimuli()))

    def get_experiment_containers(self, file_name=None,
                                  ids=None,
                                  targeted_structures=None,
                                  imaging_depths=None,
                                  cre_lines=None,
                                  transgenic_lines=None,
                                  include_failed=False,
                                  simple=True):
        """ Get a list of experiment containers matching certain criteria.

        Parameters
        ----------
        file_name: string
            File name to save/read the experiment containers.  If file_name is None,
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.

        ids: list
            List of experiment container ids.

        targeted_structures: list
            List of structure acronyms.  Must be in the list returned by
            BrainObservatoryCache.get_all_targeted_structures().

        imaging_depths: list
            List of imaging depths.  Must be in the list returned by
            BrainObservatoryCache.get_all_imaging_depths().

        cre_lines: list
            List of cre lines.  Must be in the list returned by
            BrainObservatoryCache.get_all_cre_lines().

        transgenic_lines: list
            List of transgenic lines. Must be in the list returned by
            BrainObservatoryCache.get_all_cre_lines() or.
            BrainObservatoryCache.get_all_reporter_lines().

        include_failed: boolean
            Whether or not to include failed experiment containers.

        simple: boolean
            Whether or not to simplify the dictionary properties returned by this method
            to a more concise subset.

        Returns
        -------
        list of dictionaries
        """
        _assert_not_string(targeted_structures, "targeted_structures")
        _assert_not_string(cre_lines, "cre_lines")
        _assert_not_string(transgenic_lines, "transgenic_lines")

        file_name = self.get_cache_path(
            file_name, self.EXPERIMENT_CONTAINERS_KEY)

        containers = self.api.get_experiment_containers(path=file_name,
                                                        strategy='lazy',
                                                        **Cache.cache_json())

        transgenic_lines = _merge_transgenic_lines(cre_lines, transgenic_lines)

        containers = self.api.filter_experiment_containers(containers, ids=ids,
                                                           targeted_structures=targeted_structures,
                                                           imaging_depths=imaging_depths,
                                                           transgenic_lines=transgenic_lines,
                                                           include_failed=include_failed)

        if simple:
            containers = [{
                'id': c['id'],
                'imaging_depth': c['imaging_depth'],
                'targeted_structure': c['targeted_structure']['acronym'],
                'cre_line': _find_specimen_cre_line(c['specimen']),
                'reporter_line': _find_specimen_reporter_line(c['specimen']),
                'donor_name': c['specimen']['donor']['external_donor_name'],
                'specimen_name': c['specimen']['name'],
                'tags': _find_container_tags(c),
                'failed': c['failed']
            } for c in containers]

        return containers

    def get_ophys_experiment_stimuli(self, experiment_id):
        """ For a single experiment, return the list of stimuli present in that experiment. """
        exps = self.get_ophys_experiments(ids=[experiment_id])

        if len(exps) == 0:
            return None

        return stim_info.stimuli_in_session(exps[0]['session_type'])

        
        
    def get_ophys_experiments(self, file_name=None,
                              ids=None,
                              experiment_container_ids=None,
                              targeted_structures=None,
                              imaging_depths=None,
                              cre_lines=None,
                              transgenic_lines=None,
                              stimuli=None,
                              session_types=None,
                              cell_specimen_ids=None,
                              include_failed=False,
                              simple=True):
        """ Get a list of ophys experiments matching certain criteria.

        Parameters
        ----------
        file_name: string
            File name to save/read the ophys experiments.  If file_name is None,
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.

        ids: list
            List of ophys experiment ids.

        experiment_container_ids: list
            List of experiment container ids.

        targeted_structures: list
            List of structure acronyms.  Must be in the list returned by
            BrainObservatoryCache.get_all_targeted_structures().

        imaging_depths: list
            List of imaging depths.  Must be in the list returned by
            BrainObservatoryCache.get_all_imaging_depths().

        cre_lines: list
            List of cre lines.  Must be in the list returned by
            BrainObservatoryCache.get_all_cre_lines().

        transgenic_lines: list
            List of transgenic lines. Must be in the list returned by
            BrainObservatoryCache.get_all_cre_lines() or.
            BrainObservatoryCache.get_all_reporter_lines().

        stimuli: list
            List of stimulus names.  Must be in the list returned by
            BrainObservatoryCache.get_all_stimuli().

        session_types: list
            List of stimulus session type names.  Must be in the list returned by
            BrainObservatoryCache.get_all_session_types().

        cell_specimen_ids: list
            Only include experiments that contain cells with these ids.

        include_failed: boolean
            Whether or not to include experiments from failed experiment containers.

        simple: boolean
            Whether or not to simplify the dictionary properties returned by this method
            to a more concise subset.

        Returns
        -------
        list of dictionaries
        """
        _assert_not_string(targeted_structures, "targeted_structures")
        _assert_not_string(cre_lines, "cre_lines")
        _assert_not_string(transgenic_lines, "transgenic_lines")
        _assert_not_string(stimuli, "stimuli")
        _assert_not_string(session_types, "session_types")

        file_name = self.get_cache_path(file_name, self.EXPERIMENTS_KEY)
        
        exps = self.api.get_ophys_experiments(path=file_name, 
                                              strategy='lazy', 
                                              **Cache.cache_json())

        transgenic_lines = _merge_transgenic_lines(cre_lines, transgenic_lines)

        if cell_specimen_ids is not None:
            cells = self.get_cell_specimens(ids=cell_specimen_ids)
            cell_container_ids = set([cell['experiment_container_id'] for cell in cells])
            if experiment_container_ids is not None:
                experiment_container_ids = list(set(experiment_container_ids) - cell_container_ids)
            else:
                experiment_container_ids = list(cell_container_ids)

        exps = self.api.filter_ophys_experiments(exps,
                                                 ids=ids,
                                                 experiment_container_ids=experiment_container_ids,
                                                 targeted_structures=targeted_structures,
                                                 imaging_depths=imaging_depths,
                                                 transgenic_lines=transgenic_lines,
                                                 stimuli=stimuli,
                                                 session_types=session_types,
                                                 include_failed=include_failed)

        if simple:
            exps = [{
                    'id': e['id'],
                    'imaging_depth': e['imaging_depth'],
                    'targeted_structure': e['targeted_structure']['acronym'],
                    'cre_line': _find_specimen_cre_line(e['specimen']),
                    'reporter_line': _find_specimen_reporter_line(e['specimen']),
                    'acquisition_age_days': _find_experiment_acquisition_age(e),
                    'experiment_container_id': e['experiment_container_id'],
                    'session_type': e['stimulus_name'],
                    'donor_name': e['specimen']['donor']['external_donor_name'],
                    'specimen_name': e['specimen']['name']
                    } for e in exps]
            
        return exps

    def _get_stimulus_mappings(self, file_name=None):
        """ Returns a mapping of which metrics are related to which stimuli. Internal use only. """

        file_name = self.get_cache_path(file_name, self.STIMULUS_MAPPINGS_KEY)

        mappings = self.api.get_stimulus_mappings(path=file_name,
                                                  strategy='lazy',
                                                  **Cache.cache_json())

        return mappings

    def get_cell_specimens(self,
                           file_name=None,
                           ids=None,
                           experiment_container_ids=None,
                           include_failed=False,
                           simple=True,
                           filters=None):
        """ Return cell specimens that have certain properies.

        Parameters
        ----------
        file_name: string
            File name to save/read the cell specimens.  If file_name is None,
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.

        ids: list
            List of cell specimen ids.

        experiment_container_ids: list
            List of experiment container ids.

        include_failed: bool
            Whether to include cells from failed experiment containers

        simple: boolean
            Whether or not to simplify the dictionary properties returned by this method
            to a more concise subset.

        filters: list of dicts
            List of filter dictionaries.  The Allen Brain Observatory web site can 
            generate filters in this format to reproduce a filtered set of cells
            found there.  To see what these look like, visit 
            http://observatory.brain-map.org/visualcoding, perform a cell search
            and apply some filters (e.g. find cells in a particular area), then 
            click the "view these cells in the AllenSDK" link on the bottom-left
            of the search results page.  This will take you to a page that contains
            a code sample you can use to apply those same filters via this argument.
            For more detail on the filter syntax, see BrainObservatoryApi.dataframe_query.
            

        Returns
        -------
        list of dictionaries
        """

        file_name = self.get_cache_path(file_name, self.CELL_SPECIMENS_KEY)

        cell_specimens = self.api.get_cell_metrics(path=file_name,
                                                   strategy='lazy',
                                                   **Cache.cache_json())

        cell_specimens = self.api.filter_cell_specimens(cell_specimens,
                                                        ids=ids,
                                                        experiment_container_ids=experiment_container_ids,
                                                        include_failed=include_failed,
                                                        filters=filters)

        # drop the thumbnail columns
        if simple:
            mappings = self._get_stimulus_mappings()
            thumbnails = [m['item'] for m in mappings if m[
                'item_type'] == 'T' and m['level'] == 'R']
            for cs in cell_specimens:
                for t in thumbnails:
                    del cs[t]

        return cell_specimens


    def get_ophys_experiment_data(self, ophys_experiment_id, file_name=None):
        """ Download the NWB file for an ophys_experiment (if it hasn't already been
        downloaded) and return a data accessor object.

        Parameters
        ----------
        file_name: string
            File name to save/read the data set.  If file_name is None,
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.

        ophys_experiment_id: integer
            id of the ophys_experiment to retrieve

        Returns
        -------
        BrainObservatoryNwbDataSet
        """
        file_name = self.get_cache_path(
            file_name, self.EXPERIMENT_DATA_KEY, ophys_experiment_id)

        self.api.save_ophys_experiment_data(ophys_experiment_id, file_name, strategy='lazy')

        return BrainObservatoryNwbDataSet(file_name)

    def build_manifest(self, file_name):
        """
        Construct a manifest for this Cache class and save it in a file.

        Parameters
        ----------

        file_name: string
            File location to save the manifest.

        """

        mb = ManifestBuilder()
        mb.set_version(self.MANIFEST_VERSION)
        mb.add_path('BASEDIR', '.')
        mb.add_path(self.EXPERIMENT_CONTAINERS_KEY,
                    'experiment_containers.json', typename='file', parent_key='BASEDIR')
        mb.add_path(self.EXPERIMENTS_KEY, 'ophys_experiments.json',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.EXPERIMENT_DATA_KEY, 'ophys_experiment_data/%d.nwb',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.CELL_SPECIMENS_KEY, 'cell_specimens.json',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.STIMULUS_MAPPINGS_KEY, 'stimulus_mappings.json',
                    typename='file', parent_key='BASEDIR')

        mb.write_json_file(file_name)


def _find_specimen_cre_line(specimen):
    try:
        return next(tl['name'] for tl in specimen['donor']['transgenic_lines']
                    if tl['transgenic_line_type_name'] == 'driver' and
                    'Cre' in tl['name'])
    except StopIteration:
        return None

def _find_specimen_reporter_line(specimen):
    try:
        return next(tl['name'] for tl in specimen['donor']['transgenic_lines']
                    if tl['transgenic_line_type_name'] == 'reporter')
    except StopIteration:
        return None


def _find_experiment_acquisition_age(exp):
    try:
        return (parse_date(exp['date_of_acquisition']) - parse_date(exp['specimen']['donor']['date_of_birth'])).days
    except KeyError as e:
        return None


def _merge_transgenic_lines(*lines_list):
    transgenic_lines = set()

    for lines in lines_list:
        if lines is not None:
            for line in lines:
                transgenic_lines.add(line)

    if len(transgenic_lines):
        return list(transgenic_lines)
    else:
        return None

def _find_container_tags(container):
    """ Custom logic for extracting tags from donor conditions.  Filtering 
    out tissuecyte tags. """
    conditions = container['specimen']['donor'].get('conditions', [])
    return [c['name'] for c in conditions if not c['name'].startswith('tissuecyte')]

def _assert_not_string(arg, name):
    if isinstance(arg, six.string_types):
        raise TypeError(
            "Argument '%s' with value '%s' is a string type, but should be a list." % (name, arg))



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from .rma_api import RmaApi
from jinja2 import Template


class RmaTemplate(RmaApi):
    '''
    See: `Atlas Drawings and Ontologies
    <http://help.brain-map.org/display/api/Atlas+Drawings+and+Ontologies>`_
    '''

    def __init__(self, base_uri=None, query_manifest=None):
        super(RmaTemplate, self).__init__(base_uri)
        self.templates = query_manifest

    def to_filter_rhs(self, rhs):
        if type(rhs) == list:
            return ','.join(str(r) for r in rhs)

        return rhs

    def template_query(self, template_name, entry_name, **kwargs):
        cb = self.templates[template_name]
        templates = [e for e in cb if e['name'] == entry_name]

        if len(templates) > 0:
            template = templates[0]
        else:
            raise Exception('Entry %s not found.' % (entry_name))

        query_args = {'model': template['model']}

        if 'criteria' in template:
            criteria_template = Template(template['criteria'])

            if 'criteria_params' in template:
                criteria_params = {key: self.to_filter_rhs(kwargs.get(key))
                                   for key in template['criteria_params']
                                   if key in kwargs and kwargs.get(key) is not None}
            else:
                criteria_params = {}

            criteria_str = str(criteria_template.render(**criteria_params))
            if criteria_str:
                query_args['criteria'] = criteria_str

        if 'include' in template:
            include_template = Template(template['include'])

            if 'include_params' in template:
                include_params = {key: self.to_filter_rhs(kwargs.get(key))
                                  for key in template['include_params']
                                  if key in kwargs and kwargs.get(key) is not None}
            else:
                include_params = {}

            include_str = str(include_template.render(**include_params))
            if include_str:
                query_args['include'] = include_str

        if 'only' in kwargs:
            if kwargs.get('only') is not None:
                query_args['only'] = [self.quote_string(
                    ','.join(kwargs.get('only')))]
        elif 'only' in template:
            query_args['only'] = [
                self.quote_string(','.join(template['only']))]

        if 'except' in kwargs:
            if kwargs.get('except') is not None:
                query_args['except'] = [self.quote_string(
                    ','.join(kwargs.get('except')))]
        elif 'except' in template:
            query_args['except'] = template['except']

        if 'start_row' in kwargs:
            query_args['start_row'] = kwargs.get('start_row')
        elif 'start_row' in template:
            query_args['start_row'] = template['start_row']

        if 'num_rows' in kwargs:
            query_args['num_rows'] = kwargs.get('num_rows')
        elif 'num_rows' in template:
            query_args['num_rows'] = template['num_rows']

        if 'count' in kwargs:
            query_args['count'] = kwargs.get('count')
        elif 'count' in template:
            query_args['count'] = template['count']

        if 'order' in kwargs:
            query_args['order'] = kwargs.get('order')
        elif 'order' in template:
            query_args['order'] = template['order']

        query_args.update(kwargs)

        data = self.model_query(**query_args)

        return data



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from collections import defaultdict
from six import string_types
import numpy as np
import pandas as pd

from allensdk.deprecated import class_deprecated


@class_deprecated('Use StructureTree instead.')
class Ontology(object):
    '''

    .. note:: Deprecated from 0.12.5
        `Ontology` has been replaced by `StructureTree`.

    '''

    def __init__(self, df):
        self.df = df

        child_ids = defaultdict(set)
        descendant_ids = defaultdict(set)

        for _, s in df.iterrows():
            parent_id = s['parent_structure_id']
            if np.isfinite(parent_id):
                parent_id = int(parent_id)
                child_ids[parent_id].add(s['id'])

            parent_id_list = map(int, s['structure_id_path'].split('/')[1:-1])

            for parent_id in parent_id_list:
                descendant_ids[parent_id].add(s['id'])

        self.child_ids = dict(child_ids)
        self.descendant_ids = dict(descendant_ids)

    def __getitem__(self, structures):
        """
        Return a subset of structures by id or acronym. Duplicate values are ignored.

        Parameters
        ----------

        structures: tuple
            Elements can be pandas.Series objects, which are expected to be structure ids.
            Elements can be strings, which are expected to be acronyms.
            All other elements must be cast-able to int, which are treated as structure ids.

        Returns
        -------

        pandas.DataFrame
            A subset of rows from the complete ontology that match filtering criteria.
        """

        # __getitem__ always has a single argument.  If called with a single argument
        # (e.g. ontology[315]), that item is passed straight through.  If called with
        # multiple arguments (e.g. ontology[315,997]), that gets passed through as a
        # tuple.  This normalizes the arguments so that everything is iterable.
        if not isinstance(structures, tuple) and not isinstance(structures, list) and not isinstance(structures, set):
            structures = structures,

        # this is the final set of structure ids used to filter
        structure_ids = set()

        string_strs = []
        for s in structures:
            if isinstance(s, pd.Series):
                # if it's a pandas series, assume it's a series of structure
                # ids
                structure_ids.update(s.tolist())
            elif isinstance(s, string_types):
                # if it's a string, assume it's an acronym
                string_strs.append(s)
            else:
                # if it's anything else, cast it to an integer and treat it
                # like a structure id
                structure_ids.add(int(s))

        # convert the string arguments to rows
        if len(string_strs):

            # pull out the rows that match these acronyms
            string_strs = self.df[self.df['acronym'].isin(string_strs)]

            # if there are no other structure ids, just return this dataframe
            if len(structure_ids) == 0:
                return string_strs

            # otherwise pull out the ids and add them to the set
            structure_ids.update(string_strs.id.tolist())

        return self.df.loc[structure_ids].dropna(axis=0, how='all')

    def get_descendant_ids(self, structure_ids):
        """
        Find the set of the ids of structures that are descendants of one or more structures.
        The returned set will include the input structure ids.

        Parameters
        ----------
        structure_ids: iterable
            Any iterable type that contains structure ids that can be cast to integers.

        Returns
        -------
        set
            Set of descendant structure ids.
        """

        if len(structure_ids) == 0:
            return self.descendant_ids
        else:
            descendants = set()
            for structure_id in structure_ids:
                descendants.update(self.descendant_ids.get(
                    int(structure_id), set()))
            return descendants

    def get_child_ids(self, structure_ids):
        """
        Find the set of ids that are immediate children of one or more structures.

        Parameters
        ----------
        structure_ids: iterable
            Any iterable type that contains structure ids that can be cast to integers.

        Returns
        -------
        set
            Set of child structure ids
        """

        if len(structure_ids) == 0:
            return self.child_ids
        else:
            children = set()
            for structure_id in structure_ids:
                children.update(self.child_ids.get(int(structure_id), set()))
            return children

    def get_descendants(self, structure_ids):
        """
        Find the set of structures that are descendants of one or more structures.
        The returned set will include the input structures.

        Parameters
        ----------
        structure_ids: iterable
            Any iterable type that contains structure ids that can be cast to integers.

        Returns
        -------
        pandas.DataFrame
            Set of descendant structures.
        """

        descendant_ids = self.get_descendant_ids(structure_ids)
        return self[descendant_ids]

    def get_children(self, structure_ids):
        """
        Find the set of structures that are immediate children of one or more structures.

        Parameters
        ----------
        structure_ids: iterable
            Any iterable type that contains structure ids that can be cast to integers.

        Returns
        -------
        pandas.DataFrame
            Set of child structures
        """

        child_ids = self.get_child_ids(structure_ids)
        return self[child_ids]

    def structure_descends_from(self, child_id, parent_id):
        """
        Return whether one structure id is a descendant of another structure id.
        """
        child = self[child_id]

        if child is not None:
            parent_str = '/%d/' % parent_id
            return child['structure_id_path'].values[0].find(parent_str) >= 0

        return False



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import logging

import numpy as np
import simplejson as json 
import allensdk.core.json_utilities as ju
import copy

try:
    from glif_neuron_methods import GlifNeuronMethod, METHOD_LIBRARY
except:
    from .glif_neuron_methods import GlifNeuronMethod, METHOD_LIBRARY

class GlifBadResetException( Exception ):
    """ Exception raised when voltage is still above threshold after a reset rule is applied. """
    def __init__(self, message, dv):
        super(Exception, self).__init__(message)
        self.dv = dv
            
class GlifNeuron( object ):    
    """ Implements the current-based Mihalas Neiber GLIF neuron.  Simulations model the voltage, 
    threshold, and afterspike currents of a neuron given an input stimulus.  A set of modular dynamics
    rules are applied until voltage crosses threshold, at which point a set of modular reset rules are 
    applied. See glif_neuron_methods.py for a list of what options there are for voltage, threshold, and
    afterspike current dynamics and reset rules.

    Parameters
    ----------
     El : float 
         resting potential 
     dt : float
         duration between time steps
     asc_tau_array: np.ndarray
         TODO
     R_input : float
         input resistance
     C : float
         capacitance
     asc_amp_arrap : np.ndarray
         afterspike current vector.  one element per element of asc_tau_array.
     spike_cut_length : int
         how many time steps to replace with NaNs when a spike occurs.
     th_inf : float
         instantaneous threshold
     coeffs : dict
        dictionary coefficients premultiplied to neuron properties during simulation. used for optimization.
     AScurrent_dynamics_method : dict
         dictionary containing the 'name' of the afterspike current dynamics method to use and a 'params' dictionary parameters to pass to that function.
     voltage_dynamics_method : dict
         dictionary containing the 'name' of the voltage dynamics method to use and a 'params' dictionary parameters to pass to that function.
     threshold_dynamics_method : dict
         dictionary containing the 'name' of the threshold dynamics method to use and a 'params' dictionary parameters to pass to that function.
     AScurrent_reset_method : dict
         dictionary containing the 'name' of the afterspike current dynamics method to use and a 'params' dictionary parameters to pass to that function.
     voltage_reset_method : dict
         dictionary containing the 'name' of the voltage dynamics method to use and a 'params' dictionary parameters to pass to that function.
     threshold_reset_method : dict
         dictionary containing the 'name' of the threshold dynamics method to use and a 'params' dictionary parameters to pass to that function.
     init_voltage : float 
        initial voltage value
     init_threshold : float
         initial spike threshold value
     init_AScurrents : np.ndarray
        initial afterspike current vector. one element per element of asc_tau_array.
    """

    TYPE = "GLIF"

    def __init__(self, El, dt, asc_tau_array, R_input, C, asc_amp_array, spike_cut_length, th_inf, th_adapt, coeffs,
                 AScurrent_dynamics_method, voltage_dynamics_method, threshold_dynamics_method,
                 AScurrent_reset_method, voltage_reset_method, threshold_reset_method,
                 init_voltage, init_threshold, init_AScurrents, **kwargs):

        """ Initialize the neuron."""

        self.type = GlifNeuron.TYPE
        self.El = El
        self.dt = dt
        self.asc_tau_array = np.array(asc_tau_array)
        
        self.R_input = R_input
        self.C = C

        self.asc_amp_array = np.array(asc_amp_array)
        self.spike_cut_length = int(spike_cut_length)
        self.th_inf = th_inf
        self.th_adapt = th_adapt

        self.threshold_components = None

        self.init_voltage = init_voltage
        self.init_threshold = init_threshold
        self.init_AScurrents = init_AScurrents

        assert len(asc_tau_array) == len(asc_amp_array), Exception("After-spike current vector must have same length as asc_tau_array (%d vs %d)" % (asc_amp_array, asc_tau_array))
        assert len(self.init_AScurrents) == len(self.asc_tau_array), Exception("init_AScurrents length (%d) must have same length as asc_tau_array (%d)" % (len(self.init_AScurrents), len(self.asc_tau_array)))


        # values computed based on inputs
        self.k = 1.0 / self.asc_tau_array
        self.G = 1.0 / self.R_input

        # Values that can be fit: They scale the input values.  
        # These are allowed to have default values because they are going to get optimized.
        self.coeffs = {
            'th_inf': 1,
            'C': 1,
            'G': 1,
            'b': 1,
            'a': 1,
            'asc_amp_array': np.ones(len(self.asc_tau_array))
        }

        self.coeffs.update(coeffs)
        
        logging.debug('spike cut length: %d' %  self.spike_cut_length)

        # initialize dynamics methods
        self.AScurrent_dynamics_method = self.configure_library_method('AScurrent_dynamics_method', AScurrent_dynamics_method)
        self.voltage_dynamics_method = self.configure_library_method('voltage_dynamics_method', voltage_dynamics_method)
        self.threshold_dynamics_method = self.configure_library_method('threshold_dynamics_method', threshold_dynamics_method)

        # initialize reset methods
        self.AScurrent_reset_method = self.configure_library_method('AScurrent_reset_method', AScurrent_reset_method)
        self.voltage_reset_method = self.configure_library_method('voltage_reset_method', voltage_reset_method)
        self.threshold_reset_method = self.configure_library_method('threshold_reset_method', threshold_reset_method)

    def __str__(self):
        return json.dumps(self.to_dict(), default=ju.json_handler, indent=2)

    @property
    def tau_m(self):
        return self.R_input*self.C

    @classmethod
    def from_dict(cls, d):
        return cls(El = d['El'],
                   dt = d['dt'],
                   asc_tau_array = d['asc_tau_array'],
                   R_input = d['R_input'],
                   C = d['C'],
                   asc_amp_array = d['asc_amp_array'],
                   spike_cut_length = d['spike_cut_length'],
                   th_inf = d['th_inf'],
                   th_adapt = d['th_adapt'],
                   coeffs = d.get('coeffs', {}),
                   AScurrent_dynamics_method = d['AScurrent_dynamics_method'],
                   voltage_dynamics_method = d['voltage_dynamics_method'],
                   threshold_dynamics_method = d['threshold_dynamics_method'],
                   voltage_reset_method = d['voltage_reset_method'],
                   AScurrent_reset_method = d['AScurrent_reset_method'],
                   threshold_reset_method = d['threshold_reset_method'],
                   init_voltage = d['init_voltage'],
                   init_threshold = d['init_threshold'],
                   init_AScurrents = d['init_AScurrents'])

    def to_dict(self):
        """ Convert the neuron to a serializable dictionary. """
        return {
            'type': self.type,
            'El': self.El,
            'dt': self.dt,
            'asc_tau_array': copy.deepcopy(self.asc_tau_array),
            'R_input': self.R_input,
            'C': self.C,
            'asc_amp_array': copy.deepcopy(self.asc_amp_array),
            'spike_cut_length': self.spike_cut_length,
            'th_inf': self.th_inf,
            'th_adapt': self.th_adapt,
            'coeffs': copy.deepcopy(self.coeffs),
            'AScurrent_dynamics_method': copy.deepcopy(self.AScurrent_dynamics_method),
            'voltage_dynamics_method': copy.deepcopy(self.voltage_dynamics_method),
            'threshold_dynamics_method': copy.deepcopy(self.threshold_dynamics_method),
            'AScurrent_reset_method': copy.deepcopy(self.AScurrent_reset_method),
            'voltage_reset_method': copy.deepcopy(self.voltage_reset_method),
            'threshold_reset_method': copy.deepcopy(self.threshold_reset_method),
            'init_voltage': self.init_voltage,
            'init_threshold': self.init_threshold,
            'init_AScurrents': copy.deepcopy(self.init_AScurrents), 
            'El_reference': self.El
        }

    @staticmethod
    def configure_method(method_name, method, method_params):
        """ Create a GlifNeuronMethod instance given a name, a function, and function parameters. 
        This is just a shortcut to the GlifNeuronMethod constructor.

        Parameters
        ----------
        method_name : string
            name for refering to this method later
        method : function
            a python function 
        method_parameters : dict
           function arguments whose values should be fixed

        Returns
        -------
        GlifNeuronMethod
            a GlifNeuronMethod instance
        """ 

        return GlifNeuronMethod(method_name, method, method_params)

    @staticmethod
    def configure_library_method(method_type, params):
        """ Create a GlifNeuronMethod instance out of a library of functions organized by type name. 
        This refers to the METHOD_LIBRARY in glif_neuron_methods.py, which lays out the available functions 
        that can be used for dynamics and reset rules.

        Parameters
        ----------
        method_type : string
            the name of a function category (e.g. 'AScurrent_dynamics_method' for the afterspike current dynamics methods)
        params : dict
            a dictionary with two members. 'name': the string name of function you want, and 'params': parameters you want to pass to that function

        Returns
        -------
        GlifNeuronMethod
            a GlifNeuronMethod instance
        """
        method_options = METHOD_LIBRARY.get(method_type, None)

        assert method_options is not None, Exception("Unknown method type (%s)" % method_type)
        
        method_name = params.get('name', None)
        method_params = params.get('params', None)
        
        assert method_name is not None, Exception("Method configuration for %s has no 'name'" % (method_type))
        assert method_params is not None, Exception("Method configuration for %s has no 'params'" % (method_params))
        
        method = method_options.get(method_name, None)
        
        assert method is not None, Exception("unknown method name %s of type %s" % (method_name, method_type))
        
        return GlifNeuron.configure_method(method_name, method, method_params)

    def dynamics(self, voltage_t0, threshold_t0, AScurrents_t0, inj, time_step, spike_time_steps):    
        """ Update the voltage, threshold, and afterspike currents of the neuron for a single time step.

        Parameters
        ----------
        voltage_t0 : float
            the current voltage of the neuron
        threshold_t0 : float
            the current spike threshold level of the neuron
        AScurrents_t0 : np.ndarray
            the current state of the afterspike currents in the neuron
        inj : float
            the current value of the current injection into the neuron
        time_step : int
            the current time step of the neuron simulation
        spike_time_steps : list
            a list of all of the time steps of spikes in the neuron

        Returns
        -------
        tuple
            voltage_t1 (voltage at next time step), threshold_t1 (threshold at next time step), AScurrents_t1 (afterspike currents at next time step)
        """

        AScurrents_t1 = self.AScurrent_dynamics_method(self, AScurrents_t0, time_step, spike_time_steps)
        voltage_t1 = self.voltage_dynamics_method(self, voltage_t0, AScurrents_t0, inj)
        threshold_t1 = self.threshold_dynamics_method(self, threshold_t0, voltage_t0, AScurrents_t0, inj)

        return voltage_t1, threshold_t1, AScurrents_t1
        
    def reset(self, voltage_t0, threshold_t0, AScurrents_t0):
        """ Apply reset rules to the neuron's voltage, threshold, and afterspike currents assuming a spike has occurred (voltage is above threshold). 

        Parameters
        ----------
        voltage_t0 : float
           the current voltage of the neuron
        threshold_t0 : float
            the current spike threshold level of the neuron
        AScurrents_t0 : np.ndarray
            the current state of the afterspike currents in the neuron

        Returns
        -------
        tuple
            voltage_t1 (voltage at next time step), threshold_t1 (threshold at next time step), AScurrents_t1 (afterspike currents at next time step)
        """
        
        AScurrents_t1 = self.AScurrent_reset_method(self, AScurrents_t0)  
        voltage_t1 = self.voltage_reset_method(self, voltage_t0)
        threshold_t1 = self.threshold_reset_method(self, threshold_t0, voltage_t1)
        bad_reset_flag=False
        if voltage_t1 > threshold_t1:
            bad_reset_flag=True
            #TODO put this back in eventually but would rather debug right now
#            raise GlifBadResetException("Voltage reset above threshold: voltage_t1 (%f) threshold_t1 (%f), voltage_t0 (%f) threshold_t0 (%f) AScurrents_t0 (%s)" % ( voltage_t1, threshold_t1, voltage_t0, threshold_t0, repr(AScurrents_t0)), voltage_t1 - threshold_t1)

        return voltage_t1, threshold_t1, AScurrents_t1, bad_reset_flag
    
    def run(self, stim):
        """ Run neuron simulation over a given stimulus. This steps through the stimulus applying dynamics equations.
        After each step it checks if voltage is above threshold.  If so, self.spike_cut_length NaNs are inserted 
        into the output voltages, reset rules are applied to the voltage, threshold, and afterspike currents, and the 
        simulation resumes.

        Parameters
        ----------
        stim : np.ndarray
            vector of scalar current values

        Returns
        -------
        dict
            a dictionary containing: 
                'voltage': simulated voltage values, 
                'threshold': threshold values during the simulation,
                'AScurrents': afterspike current values during the simulation, 
                'grid_spike_times': spike times (in uits of self.dt) aligned to simulation time steps, 
                'interpolated_spike_times': spike times (in units of self.dt) linearly interpolated between time steps, 
                'spike_time_steps': the indices of grid spike times, 
                'interpolated_spike_voltage': voltage of the simulation at interpolated spike times, 
                'interpolated_spike_threshold': threshold of the simulation at interpolated spike times
        """
        bad_reset_flag=False
        
        # initialize the voltage, threshold, and afterspike current values
        voltage_t0 = self.init_voltage
        threshold_t0 = self.init_threshold
        AScurrents_t0 = self.init_AScurrents

        self.threshold_components = None  #get rid of lingering method data

        num_time_steps = len(stim) 
        num_AScurrents = len(AScurrents_t0)
        
        # pre-allocate the output voltages, thresholds, and after-spike currents
        voltage_out=np.empty(num_time_steps)
        voltage_out[:]=np.nan
        threshold_out=np.empty(num_time_steps)
        threshold_out[:]=np.nan
        AScurrents_out=np.empty(shape=(num_time_steps, num_AScurrents))
        AScurrents_out[:]=np.nan        

        # array that will hold spike indices
        spike_time_steps = []
        grid_spike_times = []
        interpolated_spike_times = []
        interpolated_spike_voltage = []
        interpolated_spike_threshold = []

        time_step = 0
        while time_step < num_time_steps:
            if time_step % 10000 == 0:
                logging.info("time step %d / %d" % (time_step,  num_time_steps))

            # compute voltage, threshold, and ascurrents at current time step
            (voltage_t1, threshold_t1, AScurrents_t1) = self.dynamics(voltage_t0, threshold_t0, AScurrents_t0, stim[time_step], time_step, spike_time_steps) 

            #if the voltage is bigger than the threshold record the spike and reset the values
            if voltage_t1 > threshold_t1: 

                # spike_time_steps are stimulus indices when voltage surpassed threshold
                spike_time_steps.append(time_step)
                grid_spike_times.append(time_step * self.dt) 

                # compute higher fidelity spike time/voltage/threshold by linearly interpolating 
                interpolated_spike_times.append(interpolate_spike_time(self.dt, time_step, threshold_t0, threshold_t1, voltage_t0, voltage_t1))

                interpolated_spike_time_offset = interpolated_spike_times[-1] - (time_step - 1) * self.dt
                interpolated_spike_voltage.append(interpolate_spike_value(self.dt, interpolated_spike_time_offset, voltage_t0, voltage_t1))
                interpolated_spike_threshold.append(interpolate_spike_value(self.dt, interpolated_spike_time_offset, threshold_t0, threshold_t1))
            
                # reset voltage, threshold, and afterspike currents
                # Note that these values are not ever recorded unless the spike cut length doesnt happen (this doesnt seem quite right)
                (voltage_t0, threshold_t0, AScurrents_t0, bad_reset_flag) = self.reset(voltage_t1, threshold_t1, AScurrents_t1) 
                
                # if we are not integrating during the spike (which includes right now), insert nans then jump ahead
                # TODO MAYBE ONE LAST NAN SHOULD BE INSERTED AND THIS VALUE SHOULD BE RECORDED FOR CONSISTANCY
                if self.spike_cut_length > 0:
                    n = self.spike_cut_length

                    cut_past_end = (time_step + n) >= len(voltage_out)
                    if cut_past_end:
                        n = len(voltage_out) - time_step
                        
                    voltage_out[time_step:time_step+n] = np.nan
                    threshold_out[time_step:time_step+n] = np.nan
                    AScurrents_out[time_step:time_step+n,:] = np.nan

                    if not cut_past_end:
                        voltage_out[time_step+n] = voltage_t0 
                        threshold_out[time_step+n] = threshold_t0
                        AScurrents_out[time_step+n,:] = AScurrents_t0                    

                    time_step += self.spike_cut_length+1
                else:  
                    voltage_out[time_step] = voltage_t0 
                    threshold_out[time_step] = threshold_t0
                    AScurrents_out[time_step,:] = AScurrents_t0
                    time_step += 1
                    
                if bad_reset_flag:
                    voltage_out[time_step:time_step+5] = voltage_t0 
                    threshold_out[time_step:time_step+5] = threshold_t0
                    AScurrents_out[time_step:time_step+5] = AScurrents_t0
                    break
            else:
                # there was no spike, store the next voltages
                voltage_out[time_step] = voltage_t1 
                threshold_out[time_step] = threshold_t1
                AScurrents_out[time_step,:] = AScurrents_t1

                voltage_t0 = voltage_t1
                threshold_t0 = threshold_t1
                AScurrents_t0 = AScurrents_t1
                
                time_step += 1

        return {
            'voltage': voltage_out, 
            'threshold': threshold_out, 
            'AScurrents': AScurrents_out,
            'grid_spike_times': np.array(grid_spike_times), 
            'interpolated_spike_times': np.array(interpolated_spike_times), 
            'spike_time_steps': np.array(spike_time_steps), 
            'interpolated_spike_voltage': np.array(interpolated_spike_voltage), 
            'interpolated_spike_threshold': np.array(interpolated_spike_threshold)
            }

# TODO: DEPRICATE
#    def get_threshold_components(self):
#        if self.threshold_components is None:
#            self.threshold_components = { 'spike': [0], 'voltage': [0] }
#
#        return self.threshold_components

    def append_threshold_components(self, spike, voltage):
        self.threshold_components['spike'].append(spike)
        self.threshold_components['voltage'].append(voltage)

# TODO: DEPRICATE
#    def reset_threshold_components(self):
#        self.threshold_components = None 
            


def interpolate_spike_time(dt, time_step, threshold_t0, threshold_t1, voltage_t0, voltage_t1):
    """ Given two voltage and threshold values, the dt between them and the initial time step, interpolate
    a spike time within the dt interval by intersecting the two lines. """
    return time_step*dt + line_crossing_x(dt, voltage_t0, voltage_t1, threshold_t0, threshold_t1)


def interpolate_spike_value(dt, interpolated_spike_time_offset, v0, v1):
    """ Take a value at two adjacent time steps and linearly interpolate what the value would be
    at an offset between the two time steps. """
    return v0 + (v1 - v0) * interpolated_spike_time_offset / dt


def line_crossing_x(dx, a0, a1, b0, b1):
    """ Find the x value of the intersection of two lines. """
    assert type(a0) != int and type(a1) != int and type(b0) != int and type(b1) != int, Exception("Do not pass integers into this function!")
    return dx * (b0 - a0) / ( (a1 - a0) - (b1 - b0) )
        

def line_crossing_y(dx, a0, a1, b0, b1):
    """ Find the y value of the intersection of two lines. """
    return b0 + (b1 - b0) * (b0 - a0) / ((a1 - a0) - (b1 - b0))




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from allensdk.brain_observatory.stimulus_analysis import StimulusAnalysis
import pytest
from mock import patch, MagicMock


@pytest.fixture
def dataset():
    dataset = MagicMock(name='dataset')

    timestamps = MagicMock(name='timestamps')
    celltraces = MagicMock(name='celltraces')
    dataset.get_corrected_fluorescence_traces = \
        MagicMock(name='get_corrected_fluorescence_traces',
                  return_value=(timestamps, celltraces))
    dataset.get_roi_ids = MagicMock(name='get_roi_ids')
    dataset.get_cell_specimen_ids = MagicMock(name='get_cell_specimen_ids')
    dff_traces = MagicMock(name="dfftraces")
    dataset.get_dff_traces = MagicMock(name='get_dff_traces',
                                       return_value=(None, dff_traces))
    dxcm = MagicMock(name='dxcm')
    dxtime = MagicMock(name='dxtime')
    dataset.get_running_speed=MagicMock(name='get_running_speed',
                                        return_value=(dxcm, dxtime))
    return dataset

def mock_speed_tuning():
    binned_dx_sp = MagicMock(name='binned_dx_sp')
    binned_cells_sp = MagicMock(name='binned_cells_sp')
    binned_dx_vis = MagicMock(name='binned_dx_vis')
    binned_cells_vis = MagicMock(name='binned_cells_vis')
    peak_run = MagicMock(name='peak_run')
    
    return MagicMock(name='get_speed_tuning',
                     return_value=(binned_dx_sp,
                                   binned_cells_sp,
                                   binned_dx_vis,
                                   binned_cells_vis,
                                   peak_run))


@pytest.mark.parametrize('trigger',
                         (1,2,3,4,5))
def test_harness(dataset,
                 trigger):
    with patch('allensdk.brain_observatory.stimulus_analysis.StimulusAnalysis.get_speed_tuning',
               mock_speed_tuning()) as get_speed_tuning:
        sa = StimulusAnalysis(dataset)

        assert sa._timestamps == StimulusAnalysis._PRELOAD
        assert sa._celltraces == StimulusAnalysis._PRELOAD
        assert sa._numbercells == StimulusAnalysis._PRELOAD
        assert sa._roi_id == StimulusAnalysis._PRELOAD
        assert sa._cell_id == StimulusAnalysis._PRELOAD
        assert sa._dfftraces == StimulusAnalysis._PRELOAD
        assert sa._dxcm == StimulusAnalysis._PRELOAD
        assert sa._dxtime == StimulusAnalysis._PRELOAD

        if trigger == 1:
            print(sa.timestamps)
            print(sa.dxcm)
            print(sa.binned_dx_sp)
        elif trigger == 2:
            print(sa.celltraces)
            print(sa.dxtime)
            print(sa.binned_cells_sp)
        elif trigger == 3:
            print(sa.acquisition_rate)
            print(sa.dxcm)
            print(sa.binned_dx_vis)
        elif trigger == 4:
            print(sa.numbercells)
            print(sa.dxtime)
            print(sa.binned_cells_vis)
        elif trigger == 5:
            print(sa.timestamps)
            print(sa.dxcm)
            print(sa.peak_run)

        print(sa.roi_id)
        print(sa.cell_id)
        print(sa.dfftraces)

        dataset.get_corrected_fluorescence_traces.assert_called_once_with()
        assert sa._timestamps is not StimulusAnalysis._PRELOAD
        assert sa._celltraces is not StimulusAnalysis._PRELOAD
        assert sa._numbercells is not StimulusAnalysis._PRELOAD

        dataset.get_roi_ids.assert_called_once_with()
        assert sa._roi_id is not StimulusAnalysis._PRELOAD

        dataset.get_cell_specimen_ids.assert_called_once_with()
        assert sa._cell_id is not StimulusAnalysis._PRELOAD

        dataset.get_dff_traces.assert_called_once_with()
        assert sa._dfftraces is not StimulusAnalysis._PRELOAD

        assert sa._dxcm is not StimulusAnalysis._PRELOAD
        assert sa._dxtime is not StimulusAnalysis._PRELOAD

        get_speed_tuning.assert_called_once_with(binsize=800)
        assert sa._binned_dx_sp is not StimulusAnalysis._PRELOAD
        assert sa._binned_cells_sp is not StimulusAnalysis._PRELOAD
        assert sa._binned_dx_vis is not StimulusAnalysis._PRELOAD
        assert sa._binned_cells_vis is not StimulusAnalysis._PRELOAD
        assert sa._peak_run is not StimulusAnalysis._PRELOAD



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import patch
from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet
from allensdk.brain_observatory.session_analysis import SessionAnalysis
import os


_orig_get_stimulus_table = BrainObservatoryNwbDataSet.get_stimulus_table


def mock_stimulus_table(dset, name):
    t = _orig_get_stimulus_table(dset, name)
    t.set_value(0, 'end',
                t.loc[0,'start'] + 10)
    
    return t


@pytest.fixture
def session_a():
    filename = '/data/informatics/module_test_data/observatory/test_nwb/out_510390912.nwb'
    save_path = 'xyza'

    sa = SessionAnalysis(filename, save_path)

    return sa


@pytest.fixture
def session_b():
    filename = '/data/informatics/module_test_data/observatory/test_nwb/506278598.nwb'
    save_path = 'xyzb'

    sa = SessionAnalysis(filename, save_path)

    return sa


@pytest.fixture
def session_c():
    filename = '/data/informatics/module_test_data/observatory/test_nwb/out_510221121.nwb'
    save_path = 'xyzc'

    sa = SessionAnalysis(filename, save_path)

    return sa


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
@pytest.mark.parametrize('plot_flag',[False])
def test_session_a(session_a, plot_flag):
    with patch('allensdk.core.brain_observatory_nwb_data_set.BrainObservatoryNwbDataSet.get_stimulus_table',
               mock_stimulus_table):
        session_a.session_a(plot_flag=plot_flag)

        assert True


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
@pytest.mark.parametrize('plot_flag',[False])
def test_session_b(session_b, plot_flag):
    with patch('allensdk.core.brain_observatory_nwb_data_set.BrainObservatoryNwbDataSet.get_stimulus_table',
               mock_stimulus_table):
        session_b.session_b(plot_flag=plot_flag)

        assert True


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial_testing")
@pytest.mark.parametrize('plot_flag',[False])
def test_session_c(session_c, plot_flag):
    with patch('allensdk.core.brain_observatory_nwb_data_set.BrainObservatoryNwbDataSet.get_stimulus_table',
               mock_stimulus_table):
        session_c.session_c(plot_flag=plot_flag)

        assert True


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_session_get_session_type(session_a):
    session_type = session_a.nwb.get_session_type()

    assert session_type == 'three_session_A'


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_session_get_session_type_b(session_b):
    session_type = session_b.nwb.get_session_type()

    assert session_type == 'three_session_B'


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_session_get_session_type_c(session_c):
    session_type = session_c.nwb.get_session_type()

    assert session_type == 'three_session_C'



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np
import pandas as pd
import os
import logging


def plot_drifting_grating_traces(dg, save_dir):
    '''saves figures with a Ori X TF grid of mean resposes'''
    logging.info("Plotting Ori and TF mean response for all cells")

    blank = dg.sweep_response[dg.stim_table.temporal_frequency == 0]
    for nc in range(dg.numbercells):
        if np.mod(nc, 20) == 0:
            logging.info("Cell #%s", str(nc))
        xtime = np.arange(-1 * dg.interlength / dg.acquisition_rate, (dg.sweeplength +
                                                                      dg.interlength) / dg.acquisition_rate, 1 / dg.acquisition_rate)
        plt.figure(nc, figsize=(20, 16))
        vmax = 0
        vmin = 0
        try:
            blank_p = blank[str(nc)].mean() + \
                (blank[str(nc)].std() / len(blank[str(nc)]))
            blank_n = blank[str(nc)].mean() - \
                (blank[str(nc)].std() / len(blank[str(nc)]))
        except:
            blank_p = blank.iloc[:, nc].apply(
                np.mean) + (blank.iloc[:, nc].apply(np.std) / blank.iloc[:, nc].apply(len))
            blank_n = blank.iloc[:, nc].apply(
                np.mean) - (blank.iloc[:, nc].apply(np.std) / blank.iloc[:, nc].apply(len))
        for ori in dg.orivals:
            ori_pt = np.where(dg.orivals == ori)[0][0]
            for tf in dg.tfvals[1:]:
                tf_pt = np.where(dg.tfvals == tf)[0][0]
                sp_pt = (5 * ori_pt) + tf_pt
                subset_response = dg.sweep_response[
                    (dg.stim_table.temporal_frequency == tf) & (dg.stim_table.orientation == ori)]
                try:
                    subset_response_p = subset_response[str(nc)].mean(
                    ) + (subset_response[str(nc)][:-1].std() / len(subset_response[str(nc)]))
                    subset_response_n = subset_response[str(nc)].mean(
                    ) - (subset_response[str(nc)][:-1].std() / len(subset_response[str(nc)]))
                except:
                    subset_response_p = subset_response.iloc[:, nc].apply(
                        np.mean) + (subset_response.iloc[:, nc].apply(np.std) / subset_response.iloc[:, nc].apply(len))
                    subset_response_n = subset_response.iloc[:, nc].apply(
                        np.mean) - (subset_response.iloc[:, nc].apply(np.std) / subset_response.iloc[:, nc].apply(len))
                ax = plt.subplot(8, 5, sp_pt)
                while len(xtime) > len(subset_response[str(nc)].mean()):
                    xtime = np.delete(xtime, -1)
                try:
                    ax.fill_between(xtime, subset_response_p,
                                    subset_response_n, color='b', alpha=0.5)
                except:
                    pass
                try:
                    ax.fill_between(xtime, blank_p, blank_n,
                                    color='k', alpha=0.5)
                except:
                    pass
                try:
                    ax.plot(xtime, subset_response[
                            str(nc)].mean(), color='b', lw=2)
                except:
                    pass
                ax.plot(xtime, subset_response[
                        str(nc)].mean(), color='b', lw=2)
                # TODO: remove the [:119]  and [:-1] and the try/except
                ax.plot(xtime, blank[str(nc)].mean(), color='k', lw=2)
                ax.axvspan(0, dg.sweeplength / dg.acquisition_rate,
                           ymin=0, ymax=1, facecolor='gray', alpha=0.3)
                ax.set_xlim(-1, 3)
                ax.set_xticks(range(-1, 4))
                ax.yaxis.set_major_locator(MaxNLocator(4))
                vmax = np.where(np.amax(subset_response_p) >
                                vmax, np.amax(subset_response_p), vmax)
                vmin = np.where(np.amin(subset_response_n) <
                                vmin, np.amin(subset_response_n), vmin)

                if ori_pt < 7:
                    ax.set_xticks([])
                else:
                    ax.set_xlabel("Time (s)", fontsize=20)
                if tf_pt > 1:
                    ax.set_yticks([])
                else:
                    ax.set_ylabel(str(dg.orivals[ori_pt]), fontsize=24)
                if ori_pt == 0:
                    ax.set_title(str(dg.tfvals[tf_pt]), fontsize=24)

        for i in range(1, sp_pt + 1):
            ax = plt.subplot(8, 5, i)
            ax.set_ylim(vmin, vmax)
        plt.tick_params(labelsize=16)
        plt.tight_layout()
        plt.suptitle("Cell " + str(nc + 1), fontsize=20)
        plt.subplots_adjust(top=0.9)
        filename = 'Traces DG Cell_' + str(nc + 1) + '.png'
        fullfilename = os.path.join(save_dir, filename)
        plt.savefig(fullfilename)
        plt.close()


def plot_ns_traces(nsa, save_dir):
    logging.info("Plotting Natural Scene traces for each cell")
    xtime = np.arange(-1 * nsa.interlength / nsa.acquisition_rate, (nsa.sweeplength +
                                                                    nsa.interlength) / nsa.acquisition_rate, 1 / nsa.acquisition_rate)
    blank = nsa.sweep_response[nsa.stim_table.frame == -1]
    for nc in range(nsa.numbercells):
        if np.mod(nc, 20) == 0:
            logging.info("Cell #%s", str(nc))
        vmax = 0
        vmin = 0
        blank_p = blank[str(nc)].mean() + \
            (blank[str(nc)].std() / len(blank[str(nc)]))
        blank_n = blank[str(nc)].mean() - \
            (blank[str(nc)].std() / len(blank[str(nc)]))
        plt.figure(nc, figsize=(30, 25))
        for ns in range(nsa.number_scenes - 1):
            subset_response = nsa.sweep_response[nsa.stim_table.frame == ns]
            subset_response_p = subset_response[str(nc)].mean(
            ) + (subset_response[str(nc)][:].std() / len(subset_response[str(nc)]))
            subset_response_n = subset_response[str(nc)].mean(
            ) - (subset_response[str(nc)][:].std() / len(subset_response[str(nc)]))
            ax = plt.subplot(10, 12, ns + 1)
            try:
                ax.fill_between(xtime, subset_response_p,
                                subset_response_n, color='b', alpha=0.5)
            except:
                xtime = xtime[:-1]
                ax.fill_between(xtime, subset_response_p,
                                subset_response_n, color='b', alpha=0.5)
            ax.fill_between(xtime, blank_p, blank_n, color='k', alpha=0.5)
            ax.plot(xtime, subset_response[str(nc)].mean(), color='b', lw=2)
            ax.plot(xtime, blank[str(nc)].mean(), color='k', lw=2)
            ax.axvspan(0, nsa.sweeplength / nsa.acquisition_rate,
                       ymin=0, ymax=1, facecolor='gray', alpha=0.3)
            ax.yaxis.set_major_locator(MaxNLocator(4))
            vmax = np.where(np.amax(subset_response_p) > vmax,
                            np.amax(subset_response_p), vmax)
            vmin = np.where(np.amin(subset_response_n) < vmin,
                            np.amin(subset_response_n), vmin)
            if ns < 108:
                ax.set_xticks([])
            if np.mod(ns, 12):
                ax.set_yticks([])
        for i in range(1, nsa.number_scenes):
            ax = plt.subplot(10, 12, i)
            ax.set_ylim(vmin, vmax)
        plt.tight_layout()
        plt.suptitle("Cell " + str(nc + 1), fontsize=20)
        plt.subplots_adjust(top=0.9)
        filename = 'NS Traces Cell_' + str(nc + 1) + '.png'
        fullfilename = os.path.join(save_dir, filename)
        plt.savefig(fullfilename)
        plt.close()


def plot_sg_traces(sg, save_dir):
    logging.info("Plotting Static Grating traces for each cell")
    xtime = np.arange(-1 * sg.interlength / sg.acquisition_rate, (sg.sweeplength +
                                                                  sg.interlength) / sg.acquisition_rate, 1 / sg.acquisition_rate)
    blank = sg.sweep_response[sg.stim_table.spatial_frequency == 0]
    for nc in range(sg.numbercells):
        if np.mod(nc, 20) == 0:
            logging.info("Cell #%s", str(nc))
        vmax = 0
        vmin = 0
        blank_p = blank[str(nc)].mean() + \
            (blank[str(nc)].std() / len(blank[str(nc)]))
        blank_n = blank[str(nc)].mean() - \
            (blank[str(nc)].std() / len(blank[str(nc)]))
        while len(xtime) > len(blank_p):
            xtime = np.delete(xtime, -1)
        plt.figure(nc, figsize=(30, 30))
        ph_dict = {0: 0, 0.25: 6, 0.5: 77, 0.75: 83}
        for ori in sg.orivals:
            ori_pt = np.where(sg.orivals == ori)[0][0]
            for sf in sg.sfvals[1:]:
                sf_pt = np.where(sg.sfvals == sf)[0][0]
                for phase in sg.phasevals:
                    ph_pt = ph_dict[phase]
                    subplotnum = sf_pt + (ori_pt * 11) + ph_pt
                    subset_response = sg.sweep_response[(sg.stim_table.spatial_frequency == sf) & (
                        sg.stim_table.orientation == ori) & (sg.stim_table.phase == phase)]
                    subset_response_p = subset_response[str(nc)].mean(
                    ) + (subset_response[str(nc)][:].std() / len(subset_response[str(nc)]))
                    subset_response_n = subset_response[str(nc)].mean(
                    ) - (subset_response[str(nc)][:].std() / len(subset_response[str(nc)]))
                    ax = plt.subplot(13, 11, subplotnum)
                    ax.fill_between(xtime, subset_response_p,
                                    subset_response_n, color='b', alpha=0.5)
                    ax.fill_between(xtime, blank_p, blank_n,
                                    color='k', alpha=0.5)
                    ax.plot(xtime, subset_response[
                            str(nc)].mean(), color='b', lw=2)
                    ax.plot(xtime, blank[str(nc)].mean(), color='k', lw=2)
                    ax.axvspan(0, sg.sweeplength / sg.acquisition_rate,
                               ymin=0, ymax=1, facecolor='gray', alpha=0.3)
                    ax.yaxis.set_major_locator(MaxNLocator(4))
                    vmax = np.where(np.amax(subset_response_p)
                                    > vmax, np.amax(subset_response_p), vmax)
                    vmin = np.where(np.amin(subset_response_n)
                                    < vmin, np.amin(subset_response_n), vmin)
                    if np.mod(subplotnum, 11) != 1:
                        ax.set_yticks([])
                    else:
                        ax.set_ylabel(ori, fontsize=20)
                    if subplotnum < 133:
                        ax.set_xticks([])
                    if subplotnum < 12:
                        ax.set_title(sf, fontsize=20)
                    if subplotnum == 3:
                        ax.set_title("Phase 0.0", fontsize=20)
                    if subplotnum == 9:
                        ax.set_title("Phase 0.25", fontsize=20)
                    if subplotnum == 80:
                        ax.set_title("Phase 0.5", fontsize=20)
                    if subplotnum == 86:
                        ax.set_title("Phase 0.75", fontsize=20)
        for i in range(1, 144):
            ax = plt.subplot(13, 11, i)
            ax.set_ylim(vmin, vmax)
        plt.tight_layout()
        plt.suptitle("Cell " + str(nc + 1), fontsize=20)
        plt.subplots_adjust(top=0.9)
        filename = 'SG Traces Cell_' + str(nc + 1) + '.png'
        fullfilename = os.path.join(save_dir, filename)
        plt.savefig(fullfilename)
        plt.close()


def plot_lsn_traces(lsn, save_dir, suffix=''):
    logging.info("Plotting LSN traces for all cells")
    xtime = np.arange(-lsn.interlength / lsn.acquisition_rate,
                      (lsn.interlength + lsn.sweeplength) / lsn.acquisition_rate,
                      1.0 / lsn.acquisition_rate)

    for nc in range(lsn.numbercells):
        if np.mod(nc, 20) == 0:
            logging.info("Cell #%s", str(nc))

        plt.figure(nc, figsize=(24, 20))
        vmax = 0
        vmin = 0
        one_cell = lsn.sweep_response[str(nc)]

        for yp in range(16):
            for xp in range(28):
                sp_pt = (yp * 28) + xp + 1
                on_frame = np.where(lsn.LSN[:, yp, xp] == 255)[0]
                off_frame = np.where(lsn.LSN[:, yp, xp] == 0)[0]
                subset_on = one_cell[lsn.stim_table.frame.isin(on_frame)]
                subset_off = one_cell[lsn.stim_table.frame.isin(off_frame)]

                subset_on_mean = subset_on.mean()
                subset_off_mean = subset_off.mean()

                ax = plt.subplot(16, 28, sp_pt)
                ax.plot(xtime, subset_on_mean, color='r', lw=2)
                ax.plot(xtime, subset_off_mean, color='b', lw=2)
                ax.axvspan(0, lsn.sweeplength / lsn.acquisition_rate,
                           ymin=0, ymax=1, facecolor='gray', alpha=0.3)
                vmax = np.where(np.amax(subset_on_mean) > vmax,
                                np.amax(subset_on_mean), vmax)
                vmax = np.where(np.amax(subset_off_mean) > vmax,
                                np.amax(subset_off_mean), vmax)
                vmin = np.where(np.amin(subset_on_mean) < vmin,
                                np.amin(subset_on_mean), vmin)
                vmin = np.where(np.amin(subset_off_mean) < vmin,
                                np.amin(subset_off_mean), vmin)
                ax.set_xticks([])
                ax.set_yticks([])

        for i in range(1, sp_pt + 1):
            ax = plt.subplot(16, 28, i)
            ax.set_ylim(vmin, vmax)

        plt.tight_layout()
        plt.suptitle("Cell " + str(nc + 1), fontsize=20)
        plt.subplots_adjust(top=0.9)
        filename = 'Traces LSN Cell_' + str(nc + 1) + suffix + '.png'
        fullfilename = os.path.join(save_dir, filename)
        plt.savefig(fullfilename)
        plt.close()


def _plot_3sa(dg, nm1, nm3, save_dir):
    logging.info("Plotting for all cell")
    for nc in range(dg.numbercells):
        if np.mod(nc, 20) == 0:
            logging.info("Cell #%s", str(nc))
        plt.figure(nc, figsize=(20, 20))
        ax1 = plt.subplot2grid((6, 6), (0, 0), colspan=4)  # full trace
        ax2 = plt.subplot2grid((6, 6), (0, 4))  # histogram of F
        ax3 = plt.subplot2grid((6, 6), (1, 0), colspan=4)  # running speed
        ax4 = plt.subplot2grid((6, 6), (1, 4))
        ax5 = plt.subplot2grid((6, 6), (1, 5), sharex=ax4, sharey=ax4)
        ax6 = plt.subplot2grid((6, 6), (2, 0), colspan=2)
        ax7 = plt.subplot2grid((6, 6), (2, 2), colspan=2)
        ax8 = plt.subplot2grid((6, 6), (2, 4), colspan=2)
        ax9 = plt.subplot2grid((6, 6), (3, 0))
        ax10 = plt.subplot2grid((6, 6), (4, 0), colspan=3)
        ax11 = plt.subplot2grid((6, 6), (5, 0), colspan=3)
        ax12 = plt.subplot2grid((6, 6), (4, 3), colspan=3)
        ax13 = plt.subplot2grid((6, 6), (5, 3), colspan=3)

        xtime = np.arange(0, np.size(dg.celltraces, 1), 1.)
        xtime /= dg.acquisition_rate
        dif = np.ediff1d(dg.stim_table.start.values,
                         to_begin=8000, to_end=8000)
        test = np.argwhere(dif > 5000)
        ax1.plot(xtime, dg.celltraces[nc, :])
        for i in range(len(test) - 1):
            ax1.axvspan(xmin=(dg.stim_table.start.iloc[test[i]].values / dg.acquisition_rate), xmax=(
                dg.stim_table.end.iloc[test[i + 1] - 1].values / dg.acquisition_rate), color='gray', alpha=0.3)
        ax1.axvspan(xmin=nm1.stim_table.start.min() / nm1.acquisition_rate, xmax=(
            (nm1.stim_table.start.max() + nm1.sweeplength) / nm1.acquisition_rate), color='red', alpha=0.3)
        dif = np.ediff1d(nm3.stim_table.start.values,
                         to_begin=8000, to_end=8000)
        test = np.argwhere(dif > 5000)
        for i in range(len(test) - 1):
            ax1.axvspan(xmin=(nm3.stim_table.start.iloc[test[i]].values / nm3.acquisition_rate), xmax=(
                (nm3.stim_table.end.iloc[test[i + 1] - 1].values + nm3.sweeplength) / nm3.acquisition_rate), color='blue', alpha=0.3)
        ax1.set_xlabel("Time (s)", fontsize=20)
        ax1.set_ylabel("Fluorescence", fontsize=20)

        ax2.hist(dg.celltraces[nc, :], bins=70)
        ax2.set_yscale('log')
        ax2.set_xlabel("Fluorescence", fontsize=20)
        ax2.set_ylabel("Count", fontsize=20)

        xtime = np.arange(0, np.size(dg.dxcm), 1.)
        xtime /= dg.acquisition_rate
        ax3.plot(xtime, dg.dxcm, color='k')
        ax3.set_xlabel("Time (s)", fontsize=20)
        ax3.set_xlabel("Speed (cm/s)", fontsize=20)

        smax = nm1.binned_cells_sp[nc, np.argmax(nm1.binned_cells_sp[
                                                 nc, :, 0]), 0] + nm1.binned_cells_sp[nc, np.argmax(nm1.binned_cells_sp[nc, :, 0]), 1]
        vmax = nm1.binned_cells_vis[nc, np.argmax(nm1.binned_cells_vis[
                                                  nc, :, 0]), 0] + nm1.binned_cells_vis[nc, np.argmax(nm1.binned_cells_vis[nc, :, 0]), 1]
        rmax = np.where(smax > vmax, smax, vmax)
        smin = nm1.binned_cells_sp[nc, np.argmin(nm1.binned_cells_sp[
                                                 nc, :, 0]), 0] - nm1.binned_cells_sp[nc, np.argmin(nm1.binned_cells_sp[nc, :, 0]), 1]
        vmin = nm1.binned_cells_vis[nc, np.argmin(nm1.binned_cells_vis[
                                                  nc, :, 0]), 0] - nm1.binned_cells_vis[nc, np.argmin(nm1.binned_cells_vis[nc, :, 0]), 1]
        rmin = np.where(smin < vmin, smin, vmin)

        ax4.errorbar(nm1.binned_dx_sp[:, 0], nm1.binned_cells_sp[
                     nc, :, 0], yerr=nm1.binned_cells_sp[nc, :, 1], fmt='.', color='k')
        ax4.set_ylim(rmin, rmax)
        ax4.set_xlabel("Speed (cm/s)", fontsize=20)
        ax4.set_ylabel("DF/F", fontsize=20)
        ax4.set_title("Spontaneous", fontsize=20)

        ax5.errorbar(nm1.binned_dx_vis[:, 0], nm1.binned_cells_vis[
                     nc, :, 0], yerr=nm1.binned_cells_vis[nc, :, 1], fmt='.')
        ax5.set_ylim(rmin, rmax)
        ax5.set_xlabel("Speed (cm/s)", fontsize=20)
        ax5.set_ylabel("DF/F", fontsize=20)
        ax5.set_title("Visual Stimuli", fontsize=20)

        peakori = dg.peak.ori_dg[nc]
        peaktf = dg.peak.tf_dg[nc]
        ax6.errorbar(dg.orivals, dg.response[:, peaktf, nc, 0], yerr=dg.response[
                     :, peaktf, nc, 1], fmt='bo-', lw=2)
        ax6.fill_between(dg.orivals, np.repeat(dg.response[0, 0, nc, 0] + dg.response[0, 0, nc, 1], dg.number_ori), np.repeat(
            dg.response[0, 0, nc, 0] - dg.response[0, 0, nc, 1], dg.number_ori), color='gray', alpha=0.5)
        ax6.axhline(y=dg.response[0, 0, nc, 0], ls='--', color='k')
        ax6.annotate(str(dg.tfvals[peaktf]) + " Hz",
                     xy=(0, 0.9), xycoords='axes fraction', fontsize=14)
        ax6.set_xticks(dg.orivals)
        ax7.set_xlim(-10, 325)
        ax6.set_xlabel("Direction (d)", fontsize=20)
        ax6.set_ylabel("Mean DF/F (%)", fontsize=20)
        ax6.yaxis.set_major_locator(MaxNLocator(6))

        ax7.errorbar(range(5), dg.response[peakori, 1:, nc, 0], yerr=dg.response[
                     peakori, 1:, nc, 1], fmt='bo-', lw=2)
        ax7.fill_between(range(5), np.repeat(dg.response[0, 0, nc, 0] + dg.response[0, 0, nc, 1], 5), np.repeat(
            dg.response[0, 0, nc, 0] - dg.response[0, 0, nc, 1], 5), color='gray', alpha=0.5)
        ax7.axhline(y=dg.response[0, 0, nc, 0], ls='--', color='k', lw=2)
        ax7.annotate(str(dg.orivals[peakori]) + " Deg",
                     xy=(0, 0.9), xycoords='axes fraction', fontsize=14)
        ax7.set_xlim(-0.2, 4.2)
        ax7.set_xticks(range(5))
        ax7.set_xticklabels(dg.tfvals[1:])
        ax7.set_xlabel("Temporal frequency (Hz)", fontsize=20)

        subset = dg.sweep_response[(dg.stim_table.orientation == dg.orivals[peakori]) & (
            dg.stim_table.temporal_frequency == dg.tfvals[peaktf])]
        xtime = np.arange(-1 * dg.interlength / dg.acquisition_rate, (dg.sweeplength +
                                                                      dg.interlength) / dg.acquisition_rate, 1 / dg.acquisition_rate)
        while len(xtime) > len(subset[str(nc)].mean()):
            xtime = np.delete(xtime, -1)
        for index, row in subset.iterrows():
            ax8.plot(xtime, subset[str(nc)][index], lw=2)
        ax8.set_xlim(-1, 3)
        ax8.annotate(str(dg.orivals[peakori]) + " Deg / " + str(
            dg.tfvals[peaktf]) + " Hz", xy=(0, 0.9), xycoords='axes fraction', fontsize=14)
        ax8.set_xlabel("Time (s)", fontsize=20)
        ax8.set_ylabel("DF/F (%)", fontsize=20)
        ax8.axvspan(0, dg.sweeplength / dg.acquisition_rate,
                    ymin=0, ymax=1, facecolor='gray', alpha=0.3)
        ax8.yaxis.set_major_locator(MaxNLocator(6))
        ax8.set_title("Trial responses to prefered ori/tf", fontsize=20)

        im = ax9.imshow(dg.response[:, 1:, nc, 0],
                        cmap='gray', interpolation='none')
        ax9.set_ylabel("Direction (d)", fontsize=20)
        ax9.set_yticks(range(8))
        ax9.set_yticklabels(dg.orivals)
        ax9.set_xlabel("Temporal frequency (Hz)", fontsize=20)
        ax9.set_xticks(range(5))
        ax9.set_xticklabels(dg.tfvals[1:])
        cbar = plt.colorbar(im, ax=ax9)
        cbar.ax.set_ylabel('DF/F (%)', fontsize=8)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(8)

        xtime = np.arange(0, nm1.sweeplength /
                          nm1.acquisition_rate, 1 / nm1.acquisition_rate)
        while len(xtime) > len(nm1.sweep_response[str(nc)].mean()):
            xtime = np.delete(xtime, -1)
        for index, row in nm1.sweep_response.iterrows():
            ax10.plot(xtime, nm1.sweep_response[str(nc)][index], lw=2)
        ax10.set_xlabel("Time (s)", fontsize=20)
        ax10.set_ylabel("DF/F", fontsize=20)
        ax10.set_title("Natural Movie 1", fontsize=20, color='red')

        temp = np.empty((len(nm1.stim_table), nm1.sweeplength))
        for i in range(len(nm1.stim_table)):
            temp[i, :] = nm1.sweep_response[str(nc)].iloc[i]
        ax11.imshow(temp, cmap='gray', interpolation='none', aspect=40)
        ax11.set_ylabel("Trials", fontsize=20)
        ax11.set_xticks([])

        xtime = np.arange(0, nm3.sweeplength /
                          nm3.acquisition_rate, 1 / nm3.acquisition_rate)
        while len(xtime) > len(nm3.sweep_response[str(nc)].mean()):
            xtime = np.delete(xtime, -1)
        for index, row in nm3.sweep_response.iterrows():
            ax12.plot(xtime, nm3.sweep_response[str(nc)][index], lw=2)
        ax12.set_xlabel("Time (s)", fontsize=20)
        ax12.set_ylabel("DF/F", fontsize=20)
        ax12.set_title("Natural Movie Long", fontsize=20, color='blue')

        temp = np.empty((len(nm3.stim_table), nm3.sweeplength))
        for i in range(len(nm3.stim_table)):
            temp[i, :] = nm3.sweep_response[str(nc)].iloc[i]
        ax13.imshow(temp, cmap='gray', interpolation='none', aspect=100)
        ax13.set_ylabel("Trials", fontsize=20)
        ax13.set_xticks([])

        plt.tick_params(labelsize=16)
        plt.tight_layout()
        filename = 'Cell_' + str(nc + 1) + '_3SA.png'
        fullfilename = os.path.join(save_dir, filename)
        plt.savefig(fullfilename)
        plt.close()


def _plot_3sc(lsn, nm1, nm2, save_dir, suffix=''):
    logging.info("Plotting for all cells")
    for nc in range(lsn.numbercells):
        if np.mod(nc, 20) == 0:
            logging.info("Cell #%s", str(nc))

        plt.figure(nc, figsize=(20, 20))
        ax1 = plt.subplot2grid((6, 6), (0, 0), colspan=4)  # full trace
        ax2 = plt.subplot2grid((6, 6), (0, 4))  # histogram of F
        ax11 = plt.subplot2grid((6, 6), (1, 0), colspan=4)
        ax12 = plt.subplot2grid((6, 6), (1, 4))
        ax13 = plt.subplot2grid((6, 6), (1, 5), sharex=ax12, sharey=ax12)
        ax3 = plt.subplot2grid((6, 6), (2, 0), colspan=3)  # movie 1
        ax4 = plt.subplot2grid((6, 6), (3, 0), colspan=3)
        ax5 = plt.subplot2grid((6, 6), (2, 3), colspan=3)  # movie 2
        ax6 = plt.subplot2grid((6, 6), (3, 3), colspan=3)
        ax7 = plt.subplot2grid((6, 6), (4, 0), colspan=3)  # receptive fields
        ax8 = plt.subplot2grid((6, 6), (4, 3), colspan=3)
        ax9 = plt.subplot2grid((6, 6), (5, 0), colspan=3)
        ax10 = plt.subplot2grid((6, 6), (5, 3), colspan=3)

        xtime = np.arange(0, np.size(lsn.celltraces, 1), 1.)
        xtime /= lsn.acquisition_rate
        dif = np.ediff1d(lsn.stim_table.start.values,
                         to_begin=8000, to_end=8000)
        test = np.argwhere(dif > 5000)
        ax1.plot(xtime, lsn.celltraces[nc, :])
        for i in range(len(test) - 1):
            ax1.axvspan(xmin=(lsn.stim_table.start.iloc[test[i]].values / lsn.acquisition_rate), xmax=(
                lsn.stim_table.end.iloc[test[i + 1] - 1].values / lsn.acquisition_rate), color='gray', alpha=0.3)
        ax1.axvspan(nm1.stim_table.start.min() / nm1.acquisition_rate, nm1.stim_table.end.max() /
                    nm1.acquisition_rate, ymin=0, ymax=1, color='red', alpha=0.3)
        ax1.axvspan(nm2.stim_table.start.min() / nm2.acquisition_rate, nm2.stim_table.end.max() /
                    nm2.acquisition_rate, ymin=0, ymax=1, color='green', alpha=0.3)
        ax1.set_xlabel("Time (s)", fontsize=20)
        ax1.set_ylabel("Fluorescence", fontsize=20)

        ax2.hist(lsn.celltraces[nc, :], bins=70)
        ax2.set_yscale('log')
        ax2.set_xlabel("Fluorescence", fontsize=20)
        ax2.set_ylabel("Count", fontsize=20)

        xtime = np.arange(0, np.size(lsn.dxcm), 1.)
        xtime /= lsn.acquisition_rate
        ax11.plot(xtime, lsn.dxcm, color='k')
        ax11.set_xlabel("Time (s)", fontsize=20)
        ax11.set_xlabel("Speed (cm/s)", fontsize=20)

        smax = nm1.binned_cells_sp[nc, np.argmax(nm1.binned_cells_sp[
                                                 nc, :, 0]), 0] + nm1.binned_cells_sp[nc, np.argmax(nm1.binned_cells_sp[nc, :, 0]), 1]
        vmax = nm1.binned_cells_vis[nc, np.argmax(nm1.binned_cells_vis[
                                                  nc, :, 0]), 0] + nm1.binned_cells_vis[nc, np.argmax(nm1.binned_cells_vis[nc, :, 0]), 1]
        rmax = np.where(smax > vmax, smax, vmax)
        smin = nm1.binned_cells_sp[nc, np.argmin(nm1.binned_cells_sp[
                                                 nc, :, 0]), 0] - nm1.binned_cells_sp[nc, np.argmin(nm1.binned_cells_sp[nc, :, 0]), 1]
        vmin = nm1.binned_cells_vis[nc, np.argmin(nm1.binned_cells_vis[
                                                  nc, :, 0]), 0] - nm1.binned_cells_vis[nc, np.argmin(nm1.binned_cells_vis[nc, :, 0]), 1]
        rmin = np.where(smin < vmin, smin, vmin)

        ax12.errorbar(nm1.binned_dx_sp[:, 0], nm1.binned_cells_sp[
                      nc, :, 0], yerr=nm1.binned_cells_sp[nc, :, 1], fmt='.', color='k')
        ax12.set_ylim(rmin, rmax)
        ax12.set_xlabel("Speed (cm/s)", fontsize=20)
        ax12.set_ylabel("DF/F", fontsize=20)
        ax12.set_title("Spontaneous", fontsize=20)

        ax13.errorbar(nm1.binned_dx_vis[:, 0], nm1.binned_cells_vis[
                      nc, :, 0], yerr=nm1.binned_cells_vis[nc, :, 1], fmt='.')
        ax13.set_ylim(rmin, rmax)
        ax13.set_xlabel("Speed (cm/s)", fontsize=20)
        ax13.set_ylabel("DF/F", fontsize=20)
        ax13.set_title("Visual Stimuli", fontsize=20)

        xtime = np.arange(0, nm1.sweeplength /
                          nm1.acquisition_rate, 1 / nm1.acquisition_rate)
        for index, row in nm1.sweep_response.iterrows():
            ax3.plot(xtime, nm1.sweep_response[str(nc)][index], lw=2)
        ax3.set_xlabel("Time (s)", fontsize=20)
        ax3.set_ylabel("DF/F", fontsize=20)
        ax3.set_title("Natural Movie 1", fontsize=20, color='red')

        temp = np.empty((len(nm1.stim_table), nm1.sweeplength))
        for i in range(len(nm1.stim_table)):
            temp[i, :] = nm1.sweep_response[str(nc)].iloc[i]
        ax4.imshow(temp, cmap='gray', interpolation='none', aspect=40)
        ax4.set_ylabel("Trials", fontsize=20)
        ax4.set_xticks([])

        xtime = np.arange(0, nm2.sweeplength /
                          nm2.acquisition_rate, 1 / nm2.acquisition_rate)
        for index, row in nm2.sweep_response.iterrows():
            ax5.plot(xtime, nm2.sweep_response[str(nc)][index], lw=2)
        ax5.set_xlabel("Time (s)", fontsize=20)
        ax5.set_ylabel("DF/F", fontsize=20)
        ax5.set_title("Natural Movie 2", fontsize=20, color='green')

        temp = np.empty((len(nm2.stim_table), nm2.sweeplength))
        for i in range(len(nm2.stim_table)):
            temp[i, :] = nm2.sweep_response[str(nc)].iloc[i]
        ax6.imshow(temp, cmap='gray', interpolation='none', aspect=40)
        ax6.set_ylabel("Trials", fontsize=20)
        ax6.set_xticks([])

        vMax = np.where(np.amax(lsn.receptive_field[:, :, nc, 0]) > np.amax(lsn.receptive_field[
                        :, :, nc, 1]), np.amax(lsn.receptive_field[:, :, nc, 0]), np.amax(lsn.receptive_field[:, :, nc, 1]))
        vMin = np.where(np.amin(lsn.receptive_field[:, :, nc, 0]) < np.amin(lsn.receptive_field[
                        :, :, nc, 1]), np.amin(lsn.receptive_field[:, :, nc, 0]), np.amin(lsn.receptive_field[:, :, nc, 1]))

        imon = ax7.imshow(lsn.receptive_field[
                          :, :, nc, 0], cmap='gray', interpolation='None', vmin=vMin, vmax=vMax)
        ax7.set_title("ON", fontsize=20)
        ax7.set_xticks([])
        ax7.set_yticks([])
        cbar = plt.colorbar(imon, ax=ax7, fraction=0.046, pad=0.04)
        cbar.ax.set_ylabel('DF/F (%)', fontsize=10)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(8)

        imoff = ax8.imshow(lsn.receptive_field[
                           :, :, nc, 1], cmap='gray', interpolation='None', vmin=vMin, vmax=vMax)
        ax8.set_title("OFF", fontsize=20)
        ax8.set_xticks([])
        ax8.set_yticks([])
        cbar = plt.colorbar(imoff, ax=ax8, fraction=0.046, pad=0.04)
        cbar.ax.set_ylabel('DF/F (%)', fontsize=10)
        for t in cbar.ax.get_yticklabels():
            t.set_fontsize(8)

        zon = (lsn.receptive_field[:, :, nc, 0] - np.mean(lsn.receptive_field[
               :, :, nc, 0])) / np.std(lsn.receptive_field[:, :, nc, 0])
        zon = np.where(abs(zon) > 2, zon, 0)
        Vmax_on = np.where(abs(np.amax(zon)) > abs(
            np.amin(zon)), np.amax(zon), -1 * np.amin(zon))
        zoff = (lsn.receptive_field[:, :, nc, 1] - np.mean(lsn.receptive_field[
                :, :, nc, 1])) / np.std(lsn.receptive_field[:, :, nc, 1])
        zoff = np.where(abs(zoff) > 2, zoff, 0)
        Vmax_off = np.where(abs(np.amax(zoff)) > abs(
            np.amin(zoff)), np.amax(zoff), -1 * np.amin(zoff))
        Vmax = np.where(Vmax_on > Vmax_off, Vmax_on, Vmax_off)
        imzon = ax9.imshow(zon, cmap='RdBu_r',
                           interpolation='none', vmin=-1 * Vmax, vmax=Vmax)
        ax9.set_title("On Z-score", fontsize=20)
        ax9.set_xticks([])
        ax9.set_yticks([])
        cbar = plt.colorbar(imzon, ax=ax9, fraction=0.046, pad=0.04)

        zoff = (lsn.receptive_field[:, :, nc, 1] - np.mean(lsn.receptive_field[
                :, :, nc, 1])) / np.std(lsn.receptive_field[:, :, nc, 1])
        zoff = np.where(abs(zoff) > 2, zoff, 0)
        imzoff = ax10.imshow(
            zoff, cmap='RdBu', interpolation='none', vmin=-1 * Vmax, vmax=Vmax)
        ax10.set_title("Off Z-score", fontsize=20)
        ax10.set_xticks([])
        ax10.set_yticks([])
        cbar = plt.colorbar(imzoff, ax=ax10, fraction=0.046, pad=0.04)

        plt.tick_params(labelsize=16)
        plt.tight_layout()
        filename = 'Cell_' + str(nc + 1) + '_3SC' + suffix + '.png'
        fullfilename = os.path.join(save_dir, filename)
        plt.savefig(fullfilename)
        plt.close()


def _plot_3sb(sg, nm1, ns, save_dir):
    logging.info("Plotting for all cells")
    for nc in range(sg.numbercells):
        if np.mod(nc, 20) == 0:
            logging.info("Cell #%s", str(nc))
        plt.figure(nc, figsize=(20, 24))
        ax1 = plt.subplot2grid((6, 6), (0, 0), colspan=4)  # full trace
        ax2 = plt.subplot2grid((6, 6), (0, 4))  # histogram of F
        ax14 = plt.subplot2grid((6, 6), (1, 4))  # speed tuning
        ax16 = plt.subplot2grid((6, 6), (1, 5), sharex=ax14, sharey=ax14)
        ax15 = plt.subplot2grid((6, 6), (1, 0), colspan=4)
        ax3 = plt.subplot2grid((6, 6), (2, 0), colspan=2)  # Ori tuning
        ax4 = plt.subplot2grid((6, 6), (2, 2), colspan=2)  # sf tuning
        ax13 = plt.subplot2grid((6, 6), (2, 4), colspan=2)  # response at peak
        ax5 = plt.subplot2grid((6, 6), (3, 0))
        ax6 = plt.subplot2grid((6, 6), (3, 1))
        ax7 = plt.subplot2grid((6, 6), (3, 2))
        ax8 = plt.subplot2grid((6, 6), (3, 3))
        ax9 = plt.subplot2grid((6, 6), (4, 0), colspan=3)  # movie
        ax10 = plt.subplot2grid((6, 6), (5, 0), colspan=3)
        ax11 = plt.subplot2grid((6, 6), (4, 3), colspan=2)  # natural scenes
        ax12 = plt.subplot2grid((6, 6), (5, 3), colspan=2)

        xtime = np.arange(0, np.size(sg.celltraces, 1), 1.)
        xtime /= sg.acquisition_rate
        dif = np.ediff1d(sg.stim_table.start.values,
                         to_begin=8000, to_end=8000)
        test = np.argwhere(dif > 5000)
        ax1.plot(xtime, sg.celltraces[nc, :])
        for i in range(len(test) - 1):
            ax1.axvspan(xmin=(sg.stim_table.start.iloc[test[i]].values / sg.acquisition_rate), xmax=(
                sg.stim_table.end.iloc[test[i + 1] - 1].values / sg.acquisition_rate), color='gray', alpha=0.3)
        ax1.axvspan(nm1.stim_table.start.min() / nm1.acquisition_rate, nm1.stim_table.end.max() /
                    nm1.acquisition_rate, ymin=0, ymax=1, color='red', alpha=0.3)
        dif = np.ediff1d(ns.stim_table.start.values,
                         to_begin=8000, to_end=8000)
        test = np.argwhere(dif > 5000)
        for i in range(len(test) - 1):
            ax1.axvspan(xmin=(ns.stim_table.start.iloc[test[i]].values / ns.acquisition_rate), xmax=(
                ns.stim_table.end.iloc[test[i + 1] - 1].values / ns.acquisition_rate), color='blue', alpha=0.3)
        ax1.set_xlabel("Time (s)", fontsize=20)
        ax1.set_ylabel("Fluorescence", fontsize=20)

        ax2.hist(sg.celltraces[nc, :], bins=70)
        ax2.set_yscale('log')
        ax2.set_xlabel("Fluorescence", fontsize=20)
        ax2.set_ylabel("Count", fontsize=20)

        xtime = np.arange(0, np.size(sg.dxcm), 1.)
        xtime /= sg.acquisition_rate
        ax15.plot(xtime, sg.dxcm, color='k')
        ax15.set_xlabel("Time (s)", fontsize=20)
        ax15.set_xlabel("Speed (cm/s)", fontsize=20)

        peakori = sg.peak.ori_sg[nc]
        peaksf = sg.peak.sf_sg[nc]
        ax3.errorbar(sg.orivals, sg.response[:, peaksf, 0, nc, 0], yerr=sg.response[
                     :, peaksf, 0, nc, 1], color='blue', fmt='o-', lw=2)
        ax3.errorbar(sg.orivals, sg.response[:, peaksf, 1, nc, 0], yerr=sg.response[
                     :, peaksf, 1, nc, 1], color='cornflowerblue', fmt='o-', lw=2)
        ax3.errorbar(sg.orivals, sg.response[:, peaksf, 2, nc, 0], yerr=sg.response[
                     :, peaksf, 2, nc, 1], color='steelblue', fmt='o-', lw=2)
        ax3.errorbar(sg.orivals, sg.response[:, peaksf, 3, nc, 0], yerr=sg.response[
                     :, peaksf, 3, nc, 1], color='lightskyblue', fmt='o-', lw=2)
        ax3.fill_between(sg.orivals, np.repeat(sg.response[0, 0, 0, nc, 0] + sg.response[0, 0, 0, nc, 1], sg.number_ori), np.repeat(
            sg.response[0, 0, 0, nc, 0] - sg.response[0, 0, 0, nc, 1], sg.number_ori), color='gray', alpha=0.5)
        ax3.axhline(y=sg.response[0, 0, 0, nc, 0], ls='--', color='k', lw=2)
        ax3.set_xlim(-10, 160)
        ax3.set_xticks(sg.orivals)
        ax3.set_xlabel("Orientation (d)", fontsize=20)
        ax3.set_ylabel("DF/F (%)", fontsize=20)

        ax4.errorbar(range(5), sg.response[peakori, 1:, 0, nc, 0], yerr=sg.response[
                     peakori, 1:, 0, nc, 1], color='blue', fmt='o-', lw=2)
        ax4.errorbar(range(5), sg.response[peakori, 1:, 1, nc, 0], yerr=sg.response[
                     peakori, 1:, 1, nc, 1], color='cornflowerblue', fmt='o-', lw=2)
        ax4.errorbar(range(5), sg.response[peakori, 1:, 2, nc, 0], yerr=sg.response[
                     peakori, 1:, 2, nc, 1], color='steelblue', fmt='o-', lw=2)
        ax4.errorbar(range(5), sg.response[peakori, 1:, 3, nc, 0], yerr=sg.response[
                     peakori, 1:, 3, nc, 1], color='lightskyblue', fmt='o-', lw=2)
        ax4.fill_between(range(5), np.repeat(sg.response[0, 0, 0, nc, 0] + sg.response[0, 0, 0, nc, 1], 5), np.repeat(
            sg.response[0, 0, 0, nc, 0] - sg.response[0, 0, 0, nc, 1], 5), color='gray', alpha=0.5)
        ax4.axhline(y=sg.response[0, 0, 0, nc, 0], ls='--', color='k', lw=2)
        ax4.set_xlim(-0.2, 4.2)
        ax4.set_xticks(range(5))
        ax4.set_xticklabels(sg.sfvals[1:])
        ax4.set_xlabel("Spatial frequency (cpd)", fontsize=20)

        xtime = np.arange(-1 * sg.interlength / sg.acquisition_rate, (sg.sweeplength +
                                                                      sg.interlength) / sg.acquisition_rate, 1 / sg.acquisition_rate)
        peakori = sg.peak.ori_sg[nc]
        peaksf = sg.peak.sf_sg[nc]
        peakphase = sg.peak.phase_sg[nc]
        subset = sg.sweep_response[(sg.stim_table.orientation == sg.orivals[peakori]) & (
            sg.stim_table.spatial_frequency == sg.sfvals[peaksf]) & (sg.stim_table.phase == sg.phasevals[peakphase])]
        subset_p = subset[str(nc)].mean(
        ) + (subset[str(nc)].std() / np.sqrt(len(subset[str(nc)])))
        subset_n = subset[str(nc)].mean(
        ) - (subset[str(nc)].std() / np.sqrt(len(subset[str(nc)])))
        try:
            ax13.fill_between(xtime, subset_p, subset_n, color='b', alpha=0.5)
        except:
            xtime = xtime[:-1]
            ax13.fill_between(xtime, subset_p, subset_n, color='b', alpha=0.5)
        blank = sg.sweep_response[(sg.stim_table.orientation == 0) & (
            sg.stim_table.spatial_frequency == 0) & (sg.stim_table.phase == 0)]
        blank_p = blank[str(nc)].mean() + \
            (blank[str(nc)].std() / np.sqrt(len(blank[str(nc)])))
        blank_n = blank[str(nc)].mean() - \
            (blank[str(nc)].std() / np.sqrt(len(blank[str(nc)])))
        ax13.fill_between(xtime, blank_p, blank_n, color='gray', alpha=0.5)
        ax13.plot(xtime, subset[str(nc)].mean(), color='b', lw=2)
        ax13.plot(xtime, blank[str(nc)].mean(), color='k', lw=2)
        ax13.axvspan(0, sg.sweeplength / sg.acquisition_rate,
                     ymin=0, ymax=1, facecolor='gray', alpha=0.3)
        ax13.yaxis.set_major_locator(MaxNLocator(4))
        ax13.set_xlabel("Time (s)", fontsize=20)
        ax13.set_ylabel("DF/F (%)", fontsize=20)

        Vmax = sg.response[:, 1:, :, nc, 0].max()
        ax5.imshow(sg.response[:, 1:, 0, nc, 0], cmap='gray',
                   interpolation='none', vmin=0, vmax=Vmax)
        ax5.set_ylabel("Orientation (d)", fontsize=20)
        ax5.set_yticks(range(6))
        ax5.set_yticklabels(sg.orivals)
        ax5.set_xlabel("Spatial frequency (cpd)", fontsize=20)
        ax5.set_xticks(range(5))
        ax5.set_xticklabels(sg.sfvals[1:])
        ax5.set_title("Phase 0.0", color='blue', fontsize=20)

        ax6.imshow(sg.response[:, 1:, 1, nc, 0], cmap='gray',
                   interpolation='none', vmin=0, vmax=Vmax)
        ax6.set_xlabel("Spatial frequency (cpd)", fontsize=20)
        ax6.set_xticks(range(5))
        ax6.set_xticklabels(sg.sfvals[1:])
        ax6.set_yticks(range(6))
        ax6.set_yticklabels(sg.orivals)
        ax6.set_title("Phase 0.25", color='cornflowerblue', fontsize=20)

        ax7.imshow(sg.response[:, 1:, 2, nc, 0], cmap='gray',
                   interpolation='none', vmin=0, vmax=Vmax)
        ax7.set_xlabel("Spatial frequency (cpd)", fontsize=20)
        ax7.set_xticks(range(5))
        ax7.set_xticklabels(sg.sfvals[1:])
        ax7.set_yticks(range(6))
        ax7.set_yticklabels(sg.orivals)
        ax7.set_title("Phase 0.5", color='steelblue', fontsize=20)

        ax8.imshow(sg.response[:, 1:, 3, nc, 0], cmap='gray',
                   interpolation='none', vmin=0, vmax=Vmax)
        ax8.set_xlabel("Spatial frequency (cpd)", fontsize=20)
        ax8.set_xticks(range(5))
        ax8.set_xticklabels(sg.sfvals[1:])
        ax8.set_yticks(range(6))
        ax8.set_yticklabels(sg.orivals)
        ax8.set_title("Phase 0.75", color='lightskyblue', fontsize=20)

        xtime = np.arange(0, nm1.sweeplength /
                          nm1.acquisition_rate, 1 / nm1.acquisition_rate)
        while len(xtime) > nm1.sweeplength:
            xtime = np.delete(xtime, -1)
        for index, row in nm1.sweep_response.iterrows():
            ax9.plot(xtime, nm1.sweep_response[str(nc)][index], lw=2)
        ax9.set_xlabel("Time (s)", fontsize=20)
        ax9.set_ylabel("DF/F", fontsize=20)
        ax9.set_title("Natural Movie 1", fontsize=20, color='red')

        temp = np.empty((len(nm1.stim_table), nm1.sweeplength))
        for i in range(len(nm1.stim_table)):
            temp[i, :] = nm1.sweep_response[str(nc)].iloc[i]
        ax10.imshow(temp, cmap='gray', interpolation='none', aspect=40)
        ax10.set_ylabel("Trials", fontsize=20)
        ax10.set_xticks([])

        temp = np.copy(ns.response[1:, nc, :2])
        scene_response = pd.DataFrame(temp, columns=('response', 'error'))
        scene_response = scene_response.sort(
            columns='response', ascending=False)
        ax11.errorbar(range(ns.number_scenes - 1), scene_response.response,
                      yerr=scene_response.error, fmt='o', color='k')
        ax11.fill_between(range(ns.number_scenes - 1), np.repeat(ns.response[0, nc, 0] + ns.response[0, nc, 1], ns.number_scenes - 1), np.repeat(
            ns.response[0, nc, 0] - ns.response[0, nc, 1], ns.number_scenes - 1), color='gray', alpha=0.3)
        ax11.axhline(y=ns.response[0, nc, 0], ls='--', lw=2, color='k')
        ax11.set_xlim(-2, 120)
        ax11.set_title("Natural Scenes", fontsize=20, color='blue')
        ax11.set_xlabel("Scene", fontsize=20)
        ax11.set_ylabel("DF/F (%)", fontsize=20)

        xtime = np.arange(-1 * ns.interlength / ns.acquisition_rate, (ns.sweeplength +
                                                                      ns.interlength) / ns.acquisition_rate, 1 / ns.acquisition_rate)
        nsp = np.argmax(ns.response[1:, nc, 0])
        subset_response = ns.sweep_response[ns.stim_table.frame == nsp]
        subset_response_p = subset_response[str(nc)].mean(
        ) + (subset_response[str(nc)][:].std() / np.sqrt(len(subset_response[str(nc)])))
        subset_response_n = subset_response[str(nc)].mean(
        ) - (subset_response[str(nc)][:].std() / np.sqrt(len(subset_response[str(nc)])))
        try:
            ax12.fill_between(xtime, subset_response_p,
                              subset_response_n, color='b', alpha=0.5)
        except:
            xtime = xtime[:-1]
            ax12.fill_between(xtime, subset_response_p,
                              subset_response_n, color='b', alpha=0.5)
        blank = ns.sweep_response[ns.stim_table.frame == -1]
        blank_p = blank[str(nc)].mean() + \
            (blank[str(nc)].std() / np.sqrt(len(blank[str(nc)])))
        blank_n = blank[str(nc)].mean() - \
            (blank[str(nc)].std() / np.sqrt(len(blank[str(nc)])))
        ax12.fill_between(xtime, blank_p, blank_n, color='gray', alpha=0.5)
        ax12.plot(xtime, subset_response[str(nc)].mean(), color='b', lw=2)
        ax12.plot(xtime, blank[str(nc)].mean(), color='k', lw=2)
        ax12.axvspan(0, ns.sweeplength / ns.acquisition_rate,
                     ymin=0, ymax=1, facecolor='gray', alpha=0.3)
        ax12.yaxis.set_major_locator(MaxNLocator(4))
        ax12.set_xlabel("Time (s)", fontsize=20)
        ax12.set_ylabel("DF/F (%)", fontsize=20)

        plt.tick_params(labelsize=16)
        plt.tight_layout()
        filename = 'Cell_' + str(nc + 1) + '_3SB.png'
        fullfilename = os.path.join(save_dir, filename)
        plt.savefig(fullfilename)
        plt.close()


def plot_running_a(dg, nm1, nm3, save_dir):
    logging.info("Plotting running data summary")
    nc = -1
    plt.figure(1, figsize=(10, 8))
    ax1 = plt.subplot2grid((4, 4), (0, 0), colspan=3)
    ax2 = plt.subplot2grid((4, 4), (0, 3))
    ax3 = plt.subplot2grid((4, 4), (1, 0))
    ax4 = plt.subplot2grid((4, 4), (1, 1))
    ax5 = plt.subplot2grid((4, 4), (1, 2))
    ax6 = plt.subplot2grid((4, 4), (2, 0), colspan=2)
    ax7 = plt.subplot2grid((4, 4), (3, 0), colspan=2)
    ax8 = plt.subplot2grid((4, 4), (2, 2), colspan=2)
    ax9 = plt.subplot2grid((4, 4), (3, 2), colspan=2)

    xtime = np.arange(0, np.size(dg.dxcm), 1.)
    xtime /= dg.acquisition_rate
    dif = np.ediff1d(dg.stim_table.start.values, to_begin=8000, to_end=8000)
    test = np.argwhere(dif > 5000)
    ax1.plot(xtime, dg.dxcm, color='k')
    for i in range(len(test) - 1):
        ax1.axvspan(xmin=(dg.stim_table.start.iloc[test[i]].values / dg.acquisition_rate), xmax=(
            dg.stim_table.end.iloc[test[i + 1] - 1].values / dg.acquisition_rate), color='gray', alpha=0.3)
    ax1.axvspan(xmin=nm1.stim_table.start.min() / nm1.acquisition_rate, xmax=(
        (nm1.stim_table.start.max() + nm1.sweeplength) / nm1.acquisition_rate), color='red', alpha=0.3)
    dif = np.ediff1d(nm3.stim_table.start.values, to_begin=8000, to_end=8000)
    test = np.argwhere(dif > 5000)
    for i in range(len(test) - 1):
        ax1.axvspan(xmin=(nm3.stim_table.start.iloc[test[i]].values / nm3.acquisition_rate), xmax=(
            (nm3.stim_table.end.iloc[test[i + 1] - 1].values + nm3.sweeplength) / nm3.acquisition_rate), color='blue', alpha=0.3)
    ax1.set_xlabel("Time (s)", fontsize=20)
    ax1.set_ylabel("Speed (cm/s)", fontsize=20)

    dx = dg.dxcm[np.logical_not(np.isnan(dg.dxcm))]
    ax2.hist(dx, bins=80, range=(-20, 100), color='gray')
    ax2.set_xlabel("Speed (cm/s)", fontsize=20)

    run_peak = np.where(dg.response[:, 1:, nc, 0]
                        == np.nanmax(dg.response[:, 1:, nc, 0]))
    peakori = run_peak[0][0]
    peaktf = run_peak[1][0] + 1

    ax3.errorbar(dg.orivals, dg.response[:, peaktf, nc, 0], yerr=dg.response[
                 :, peaktf, nc, 1], fmt='b.-', lw=2)
    ax3.fill_between(dg.orivals, np.repeat(dg.response[0, 0, nc, 0] + dg.response[0, 0, nc, 1], dg.number_ori), np.repeat(
        dg.response[0, 0, nc, 0] - dg.response[0, 0, nc, 1], dg.number_ori), color='gray', alpha=0.5)
    ax3.axhline(y=dg.response[0, 0, nc, 0], ls='--', color='k')
    ax3.annotate(str(dg.tfvals[peaktf]) + " Hz",
                 xy=(0, 0.9), xycoords='axes fraction', fontsize=14)
    ax3.set_xtick = (dg.orivals)
    ax3.set_xlabel("Direction (deg)", fontsize=20)
    ax3.set_ylabel("Speed (cm/s)", fontsize=20)
    ax3.yaxis.set_major_locator(MaxNLocator(6))

    ax4.errorbar(dg.tfvals[1:], dg.response[peakori, 1:, nc, 0], yerr=dg.response[
                 peakori, 1:, nc, 1], fmt='b.-', lw=2)
    ax4.fill_between(dg.tfvals[1:], np.repeat(dg.response[0, 0, nc, 0] + dg.response[0, 0, nc, 1], dg.number_tf - 1),
                     np.repeat(dg.response[0, 0, nc, 0] - dg.response[0, 0, nc, 1], dg.number_tf - 1), color='gray', alpha=0.5)
    ax4.axhline(y=dg.response[0, 0, nc, 0], ls='--', color='k')
    ax4.annotate(str(dg.orivals[peakori]) + " Deg",
                 xy=(0, 0.9), xycoords='axes fraction', fontsize=14)
    ax4.set_xticks = (dg.tfvals[1:])
    ax4.set_xlabel("Temporal frequency (Hz)", fontsize=20)
    ax4.yaxis.set_major_locator(MaxNLocator(6))

    im = ax5.imshow(dg.response[:, 1:, nc, 0],
                    cmap='gray', interpolation='none')
    ax5.set_ylabel("Direction", fontsize=16)
    ax5.set_xlabel("TF", fontsize=16)
    ax5.set_yticks(range(dg.number_ori))
    ax5.set_yticklabels(list(dg.orivals.astype(int).astype(str)))
    ax5.set_xticks(range(dg.number_tf - 1))
    ax5.set_xticklabels(list(dg.tfvals[1:].astype(int).astype(str)))
    cbar = plt.colorbar(im, ax=ax5)
    cbar.ax.set_ylabel('Speed (cm/s)', fontsize=8)
    for t in cbar.ax.get_yticklabels():
        t.set_fontsize(8)

    xtime = np.arange(0, nm1.sweeplength /
                      nm1.acquisition_rate, 1 / nm1.acquisition_rate)
    while len(xtime) > len(nm1.sweep_response['dx'].mean()):
        xtime = np.delete(xtime, -1)
    for index, row in nm1.sweep_response.iterrows():
        ax6.plot(xtime, nm1.sweep_response['dx'][index], lw=2)
    ax6.set_xlabel("Time (s)", fontsize=20)
    ax6.set_ylabel("DF/F", fontsize=20)
    ax6.set_title("Natural Movie 1", fontsize=20, color='red')

    temp = np.empty((len(nm1.stim_table), nm1.sweeplength))
    for i in range(len(nm1.stim_table)):
        temp[i, :] = nm1.sweep_response['dx'].iloc[i][:, 0]
    ax7.imshow(temp, cmap='gray', interpolation='none', aspect=40)
    ax7.set_ylabel("Trials", fontsize=20)
    ax7.set_xticks([])

    xtime = np.arange(0, nm3.sweeplength /
                      nm3.acquisition_rate, 1 / nm3.acquisition_rate)
    while len(xtime) > len(nm3.sweep_response['dx'].mean()):
        xtime = np.delete(xtime, -1)
    for index, row in nm3.sweep_response.iterrows():
        ax8.plot(xtime, nm3.sweep_response['dx'][index], lw=2)
    ax8.set_xlabel("Time (s)", fontsize=20)
    ax8.set_ylabel("DF/F", fontsize=20)
    ax8.set_title("Natural Movie Long", fontsize=20, color='blue')

    temp = np.empty((len(nm3.stim_table), nm3.sweeplength))
    for i in range(len(nm3.stim_table)):
        temp[i, :] = nm3.sweep_response['dx'].iloc[i][:, 0]
    ax9.imshow(temp, cmap='gray', interpolation='none', aspect=100)
    ax9.set_ylabel("Trials", fontsize=20)
    ax9.set_xticks([])

    plt.tick_params(labelsize=16)
    plt.tight_layout()
    plt.suptitle("Running Summary", fontsize=20)
    plt.subplots_adjust(top=0.9)
    filename = 'Running Summary.png'
    fullfilename = os.path.join(save_dir, filename)
    plt.savefig(fullfilename)
    plt.close()



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from .utilities import smooth
import numpy as np
import scipy.stats as sps

def detect_events(data, cell_index, stimulus, debug_plots=False):



    stimulus_table = data.get_stimulus_table(stimulus)
    dff_trace = data.get_dff_traces()[1][cell_index, :]



    k_min = 0
    k_max = 10
    delta = 3

    dff_trace = smooth(dff_trace, 5)



    var_dict = {}
    debug_dict = {}
    for ii, fi in enumerate(stimulus_table['start'].values):

        if ii > 0 and stimulus_table.iloc[ii].start == stimulus_table.iloc[ii-1].end:
            offset = 1
        else:
            offset = 0

        if fi + k_min >= 0 and fi + k_max <= len(dff_trace):
            trace = dff_trace[fi + k_min+1+offset:fi + k_max+1+offset]

            xx = (trace - trace[0])[delta] - (trace - trace[0])[0]
            yy = max((trace - trace[0])[delta + 2] - (trace - trace[0])[0 + 2],
                     (trace - trace[0])[delta + 3] - (trace - trace[0])[0 + 3],
                     (trace - trace[0])[delta + 4] - (trace - trace[0])[0 + 4])

            var_dict[ii] = (trace[0], trace[-1], xx, yy)
            debug_dict[fi + k_min+1+offset] = (ii, trace)

    xx_list, yy_list = [], []
    for _, _, xx, yy in var_dict.itervalues():
        xx_list.append(xx)
        yy_list.append(yy)

    mu_x = np.median(xx_list)
    mu_y = np.median(yy_list)

    xx_centered = np.array(xx_list)-mu_x
    yy_centered = np.array(yy_list)-mu_y

    std_factor = 1
    std_x = 1./std_factor*np.percentile(np.abs(xx_centered), [100*(1-2*(1-sps.norm.cdf(std_factor)))])
    std_y = 1./std_factor*np.percentile(np.abs(yy_centered), [100*(1-2*(1-sps.norm.cdf(std_factor)))])

    curr_inds = []
    allowed_sigma = 4
    for ii, (xi, yi) in enumerate(zip(xx_centered, yy_centered)):
        if np.sqrt(((xi)/std_x)**2+((yi)/std_y)**2) < allowed_sigma:
            curr_inds.append(True)
        else:
            curr_inds.append(False)

    curr_inds = np.array(curr_inds)
    data_x = xx_centered[curr_inds]
    data_y = yy_centered[curr_inds]
    Cov = np.cov(data_x, data_y)
    Cov_Factor = np.linalg.cholesky(Cov)
    Cov_Factor_Inv = np.linalg.inv(Cov_Factor)

    #===================================================================================================================

    noise_threshold = max(allowed_sigma * std_x + mu_x, allowed_sigma * std_y + mu_y)
    mu_array = np.array([mu_x, mu_y])
    yes_set, no_set = set(), set()
    for ii, (t0, tf, xx, yy) in var_dict.iteritems():


        xi_z, yi_z = Cov_Factor_Inv.dot((np.array([xx,yy]) - mu_array))

        # Conditions in order:
        # 1) Outside noise blob
        # 2) Minimum change in df/f
        # 3) Change evoked by this trial, not previous
        # 4) At end of trace, ended up outside of noise floor

        if np.sqrt(xi_z**2 + yi_z**2) > 4 and yy > .05 and xx < yy and tf > noise_threshold/2:
            yes_set.add(ii)
        else:
            no_set.add(ii)



    assert len(var_dict) == len(stimulus_table)
    b = np.zeros(len(stimulus_table), dtype=np.bool)
    for yi in yes_set:
        b[yi] = True

    if debug_plots == True:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1,2)
        # ax[0].plot(dff_trace)
        for key, val in debug_dict.iteritems():
            ti, trace = val
            if ti in no_set:
                ax[0].plot(np.arange(key, key+len(trace)), trace, 'b')
            elif ti in yes_set:
                ax[0].plot(np.arange(key, key + len(trace)), trace, 'r', linewidth=2)
            else:
                raise Exception

        for ii in yes_set:
            ax[1].plot([var_dict[ii][2]], [var_dict[ii][3]], 'r.')

        for ii in no_set:
            ax[1].plot([var_dict[ii][2]], [var_dict[ii][3]], 'b.')

        print('number_of_events:', b.sum())
        plt.show()

    return b
